from sound_lib.main import bass_call
import ctypes
from sound_lib.external import pybass
import string  # for the alphabet!


class SoundEffect(object):
    """ """
    def __init__(self, channel, type=None, priority=0):
        self.original_channel = channel
        if hasattr(channel, "handle"):
            channel = channel.handle
        if type is None:
            type = self.effect_type
        self.effect_type = type
        self.priority = priority
        self.handle = bass_call(pybass.BASS_ChannelSetFX, channel, type, priority)

    def get_parameters(self):
        """Retrieves the parameters of an effect."""
        res = {}
        params = self.struct()
        bass_call(pybass.BASS_FXGetParameters, self.handle, ctypes.pointer(params))
        for f in params._fields_:
            res[f[0]] = getattr(params, f[0])
        return res

    def set_parameters(self, parameters):
        """

        Args:
          parameters: 

        Returns:

        """
        params = self.struct()
        for p, v in parameters.items():
            setattr(params, p, v)
        bass_call(pybass.BASS_FXSetParameters, self.handle, ctypes.pointer(params))

    def __dir__(self):
        res = dir(self.__class__)
        return res + self._get_pythonic_effect_fields()

    def _get_effect_fields(self):
        """ """
        return [i[0] for i in self.struct._fields_]

    def _get_pythonic_effect_fields(self):
        """ """
        return [
            self._bass_to_python(i)
            for i in self._get_effect_fields()
            if not i.startswith("_")
        ]

    def _bass_to_python(self, func):
        """

        Args:
          func: 

        Returns:

        """
        for c in string.ascii_lowercase:
            func = func.replace(c.upper(), "_%s" % c)
        if func.startswith("_"):
            func = func[1:]
        return func[2:]

    def _python_to_bass(self, func):
        """

        Args:
          func: 

        Returns:

        """
        for c in string.ascii_lowercase:
            func = func.replace("_%s" % c, c.upper())
        func = "%s%s" % (func[0].upper(), func[1:])
        for f in self._get_effect_fields():
            if f.endswith(func):
                func = f
        return func

    def __getattr__(self, attr):
        return self.get_parameters()[self._python_to_bass(attr)]

    def __setattr__(self, attr, val):
        if attr not in self._get_pythonic_effect_fields():
            return super(SoundEffect, self).__setattr__(attr, val)
        params = self.get_parameters()
        key = self._python_to_bass(attr)
        if key not in params:
            raise AttributeError(
                "Unable to set attribute, suspect issue with base name-munging code"
            )
        params[key] = val
        self.set_parameters(params)



from __future__ import absolute_import
# Copyright(c) Max Kolosov 2009 maxkolosov@inbox.ru
# http://vosolok2008.narod.ru
# BSD license

__version__ = '0.1'
__versionTime__ = '2009-11-15'
__author__ = 'Max Kolosov <maxkolosov@inbox.ru>'
__doc__ = '''
pybassmix.py - is ctypes python module for
BASSmix - extension to the BASS audio library, providing the ability
to mix together multiple BASS channels, with resampling and matrix mixing
features. It also provides the ability to go the other way and split a
BASS channel into multiple channels.
'''

import os, sys, ctypes, platform, pybass

QWORD = pybass.QWORD
HSYNC = pybass.HSYNC
HSTREAM = pybass.HSTREAM
DOWNLOADPROC = pybass.DOWNLOADPROC
SYNCPROC = pybass.SYNCPROC
BASS_FILEPROCS = pybass.BASS_FILEPROCS

from .paths import x86_path, x64_path
import libloader

bassmix_module = libloader.load_library('bassmix', x86_path=x86_path, x64_path=x64_path)
func_type = libloader.get_functype()
	
# additional BASS_SetConfig option
BASS_CONFIG_MIXER_FILTER = 0x10600
BASS_CONFIG_MIXER_BUFFER = 0x10601
BASS_CONFIG_SPLIT_BUFFER = 0x10610

# BASS_Mixer_StreamCreate flags
BASS_MIXER_END = 0x10000# end the stream when there are no sources
BASS_MIXER_NONSTOP = 0x20000# don't stall when there are no sources
BASS_MIXER_RESUME = 0x1000# resume stalled immediately upon new/unpaused source

# source flags
BASS_MIXER_FILTER = 0x1000# resampling filter
BASS_MIXER_BUFFER = 0x2000# buffer data for BASS_Mixer_ChannelGetData/Level
BASS_MIXER_LIMIT = 0x4000# limit mixer processing to the amount available from this source
BASS_MIXER_MATRIX = 0x10000# matrix mixing
BASS_MIXER_PAUSE = 0x20000# don't process the source
BASS_MIXER_DOWNMIX = 0x400000# downmix to stereo/mono
BASS_MIXER_NORAMPIN = 0x800000# don't ramp-in the start

# envelope node
class BASS_MIXER_NODE(ctypes.Structure):
	_fields_ = [('pos', ctypes.c_ulong),#QWORD pos;
				('value', ctypes.c_float)#float value;
				]

# envelope types
BASS_MIXER_ENV_FREQ = 1
BASS_MIXER_ENV_VOL = 2
BASS_MIXER_ENV_PAN = 3
BASS_MIXER_ENV_LOOP = 0x10000# FLAG: loop

# additional sync type
BASS_SYNC_MIXER_ENVELOPE = 0x10200

# BASS_CHANNELINFO type
BASS_CTYPE_STREAM_MIXER = 0x10800
BASS_CTYPE_STREAM_SPLIT = 0x10801


#DWORD BASSMIXDEF(BASS_Mixer_GetVersion)();
BASS_Mixer_GetVersion = func_type(ctypes.c_ulong)(('BASS_Mixer_GetVersion', bassmix_module))

#HSTREAM BASSMIXDEF(BASS_Mixer_StreamCreate)(DWORD freq, DWORD chans, DWORD flags);
BASS_Mixer_StreamCreate = func_type(HSTREAM, ctypes.c_ulong, ctypes.c_ulong, ctypes.c_ulong)(('BASS_Mixer_StreamCreate', bassmix_module))
#BOOL BASSMIXDEF(BASS_Mixer_StreamAddChannel)(HSTREAM handle, DWORD channel, DWORD flags);
BASS_Mixer_StreamAddChannel = func_type(ctypes.c_byte, HSTREAM, ctypes.c_ulong, ctypes.c_ulong)(('BASS_Mixer_StreamAddChannel', bassmix_module))
#BOOL BASSMIXDEF(BASS_Mixer_StreamAddChannelEx)(HSTREAM handle, DWORD channel, DWORD flags, QWORD start, QWORD length);
BASS_Mixer_StreamAddChannelEx = func_type(ctypes.c_byte, HSTREAM, ctypes.c_ulong, ctypes.c_ulong, QWORD, QWORD)(('BASS_Mixer_StreamAddChannelEx', bassmix_module))

#HSTREAM BASSMIXDEF(BASS_Mixer_ChannelGetMixer)(DWORD handle);
BASS_Mixer_ChannelGetMixer = func_type(HSTREAM, ctypes.c_ulong)(('BASS_Mixer_ChannelGetMixer', bassmix_module))
#DWORD BASSMIXDEF(BASS_Mixer_ChannelFlags)(DWORD handle, DWORD flags, DWORD mask);
BASS_Mixer_ChannelFlags = func_type(ctypes.c_ulong, ctypes.c_ulong, ctypes.c_ulong, ctypes.c_ulong)(('BASS_Mixer_ChannelFlags', bassmix_module))
#BOOL BASSMIXDEF(BASS_Mixer_ChannelRemove)(DWORD handle);
BASS_Mixer_ChannelRemove = func_type(ctypes.c_byte, ctypes.c_ulong)(('BASS_Mixer_ChannelRemove', bassmix_module))
#BOOL BASSMIXDEF(BASS_Mixer_ChannelSetPosition)(DWORD handle, QWORD pos, DWORD mode);
BASS_Mixer_ChannelSetPosition = func_type(ctypes.c_byte, ctypes.c_ulong, QWORD, ctypes.c_ulong)(('BASS_Mixer_ChannelSetPosition', bassmix_module))
#QWORD BASSMIXDEF(BASS_Mixer_ChannelGetPosition)(DWORD handle, DWORD mode);
BASS_Mixer_ChannelGetPosition = func_type(QWORD, ctypes.c_ulong, ctypes.c_ulong)(('BASS_Mixer_ChannelGetPosition', bassmix_module))
#DWORD BASSMIXDEF(BASS_Mixer_ChannelGetLevel)(DWORD handle);
BASS_Mixer_ChannelGetLevel = func_type(ctypes.c_ulong, ctypes.c_ulong)(('BASS_Mixer_ChannelGetLevel', bassmix_module))
#DWORD BASSMIXDEF(BASS_Mixer_ChannelGetData)(DWORD handle, void *buffer, DWORD length);
BASS_Mixer_ChannelGetData = func_type(ctypes.c_ulong, ctypes.c_ulong, ctypes.c_void_p, ctypes.c_ulong)(('BASS_Mixer_ChannelGetData', bassmix_module))
#HSYNC BASSMIXDEF(BASS_Mixer_ChannelSetSync)(DWORD handle, DWORD type, QWORD param, SYNCPROC *proc, void *user);
BASS_Mixer_ChannelSetSync = func_type(HSYNC, ctypes.c_ulong, ctypes.c_ulong, QWORD, SYNCPROC, ctypes.c_void_p)(('BASS_Mixer_ChannelSetSync', bassmix_module))
#BOOL BASSMIXDEF(BASS_Mixer_ChannelRemoveSync)(DWORD channel, HSYNC sync);
BASS_Mixer_ChannelRemoveSync = func_type(ctypes.c_byte, ctypes.c_ulong, HSYNC)(('BASS_Mixer_ChannelRemoveSync', bassmix_module))
#BOOL BASSMIXDEF(BASS_Mixer_ChannelSetMatrix)(DWORD handle, const float *matrix);
BASS_Mixer_ChannelSetMatrix = func_type(ctypes.c_byte, ctypes.c_ulong, ctypes.POINTER(ctypes.c_float))(('BASS_Mixer_ChannelSetMatrix', bassmix_module))
#BOOL BASSMIXDEF(BASS_Mixer_ChannelGetMatrix)(DWORD handle, float *matrix);
BASS_Mixer_ChannelGetMatrix = func_type(ctypes.c_byte, ctypes.c_ulong, ctypes.POINTER(ctypes.c_float))(('BASS_Mixer_ChannelGetMatrix', bassmix_module))
#BOOL BASSMIXDEF(BASS_Mixer_ChannelSetEnvelope)(DWORD handle, DWORD type, const BASS_MIXER_NODE *nodes, DWORD count);
BASS_Mixer_ChannelSetEnvelope = func_type(ctypes.c_byte, ctypes.c_ulong, ctypes.c_ulong, ctypes.POINTER(BASS_MIXER_NODE), ctypes.c_ulong)(('BASS_Mixer_ChannelSetEnvelope', bassmix_module))
#BOOL BASSMIXDEF(BASS_Mixer_ChannelSetEnvelopePos)(DWORD handle, DWORD type, QWORD pos);
BASS_Mixer_ChannelSetEnvelopePos = func_type(ctypes.c_byte, ctypes.c_ulong, ctypes.c_ulong, QWORD)(('BASS_Mixer_ChannelSetEnvelopePos', bassmix_module))
#QWORD BASSMIXDEF(BASS_Mixer_ChannelGetEnvelopePos)(DWORD handle, DWORD type, float *value);
BASS_Mixer_ChannelGetEnvelopePos = func_type(QWORD, ctypes.c_ulong, ctypes.c_ulong, ctypes.c_float)(('BASS_Mixer_ChannelGetEnvelopePos', bassmix_module))

#HSTREAM BASSMIXDEF(BASS_Split_StreamCreate)(DWORD channel, DWORD flags, int *chanmap);
BASS_Split_StreamCreate = func_type(HSTREAM, ctypes.c_ulong, ctypes.c_ulong, ctypes.c_int)(('BASS_Split_StreamCreate', bassmix_module))
#DWORD BASSMIXDEF(BASS_Split_StreamGetSource)(HSTREAM handle);
BASS_Split_StreamGetSource = func_type(ctypes.c_ulong, HSTREAM)(('BASS_Split_StreamGetSource', bassmix_module))
#BOOL BASSMIXDEF(BASS_Split_StreamReset)(DWORD handle);
BASS_Split_StreamReset = func_type(ctypes.c_byte, ctypes.c_ulong)(('BASS_Split_StreamReset', bassmix_module))




from __future__ import absolute_import
"BASS_FX wrapper by Christopher Toth"""

import ctypes
import os
from . import pybass
from .paths import x86_path, x64_path
import libloader

bass_fx_module = libloader.load_library('bass_fx', x86_path=x86_path, x64_path=x64_path)
func_type = libloader.get_functype()


#Error codes returned by BASS_ErrorGetCode
BASS_ERROR_FX_NODECODE = 4000
BASS_ERROR_FX_BPMINUSE = 4001

#Tempo / Reverse / BPM / Beat flag
BASS_FX_FREESOURCE = 0x10000

#BASS_FX Version 
BASS_FX_GetVersion = func_type(ctypes.c_ulong)(('BASS_FX_GetVersion', bass_fx_module))


"""D S P (Digital Signal Processing)"""

"""
Multi-channel order of each channel is as follows:
	 3 channels       left-front, right-front, center.
	 4 channels       left-front, right-front, left-rear/side, right-rear/side.
	 6 channels (5.1) left-front, right-front, center, LFE, left-rear/side, right-rear/side.
	 8 channels (7.1) left-front, right-front, center, LFE, left-rear/side, right-rear/side, left-rear center, right-rear center.
"""

#DSP channels flags
BASS_BFX_CHANALL = -1  #all channels at once (as by default)
BASS_BFX_CHANNONE = 0 #disable an effect for all channels
BASS_BFX_CHAN1 = 1 #left-front channel
BASS_BFX_CHAN2 = 2 #right-front channel
BASS_BFX_CHAN3 = 4 #see above info
BASS_BFX_CHAN4 = 8 #see above info
BASS_BFX_CHAN5 = 16
BASS_BFX_CHAN6 = 32
BASS_BFX_CHAN7 = 64
BASS_BFX_CHAN8 = 128

#DSP effects
(
 BASS_FX_BFX_ROTATE,
 BASS_FX_BFX_ECHO,
 BASS_FX_BFX_FLANGER,
 BASS_FX_BFX_VOLUME,
 BASS_FX_BFX_PEAKEQ,
 BASS_FX_BFX_REVERB,
 BASS_FX_BFX_LPF,
 BASS_FX_BFX_MIX,
 BASS_FX_BFX_DAMP,
 BASS_FX_BFX_AUTOWAH,
 BASS_FX_BFX_ECHO2,
 BASS_FX_BFX_PHASER,
 BASS_FX_BFX_ECHO3,
 BASS_FX_BFX_CHORUS,
 BASS_FX_BFX_APF,
 BASS_FX_BFX_COMPRESSOR,
 BASS_FX_BFX_DISTORTION,
 BASS_FX_BFX_COMPRESSOR2,
 BASS_FX_BFX_VOLUME_ENV,
 BASS_FX_BFX_BQF,
) = range(0x10000, 0x10000+20)

#BiQuad filters
(
 BASS_BFX_BQF_LOWPASS,
 BASS_BFX_BQF_HIGHPASS,
 BASS_BFX_BQF_BANDPASS, #constant 0 dB peak gain
 BASS_BFX_BQF_BANDPASS_Q, #constant skirt gain, peak gain = Q
 BASS_BFX_BQF_NOTCH,
 BASS_BFX_BQF_ALLPASS,
 BASS_BFX_BQF_PEAKINGEQ,
 BASS_BFX_BQF_LOWSHELF,
 BASS_BFX_BQF_HIGHSHELF,
) = range(9)

#Echo
class BASS_BFX_ECHO(ctypes.Structure):
 _fields_ = [
  ('fLevel', ctypes.c_float), #[0....1....n] linear
  ('lDelay', ctypes.c_int), #[1200..30000]
 ]

#Flanger
class BASS_BFX_FLANGER(ctypes.Structure):
 _fields_ = [
  ('fWetDry', ctypes.c_float), #[0....1....n] linear
  ('fSpeed', ctypes.c_float), #[0......0.09]
  ('lChannel', ctypes.c_int), #BASS_BFX_CHANxxx flag/s
 ]

#volume
class BASS_BFX_VOLUME(ctypes.Structure):
 _fields_ = [
  ('lChannel', ctypes.c_int), #BASS_BFX_CHANxxx flag/s or 0 for global volume control
  ('fVolume', ctypes.c_float), #[0....1....n] linear
 ]

#Peaking Equalizer
class BASS_BFX_PEAKEQ(ctypes.Structure):
 _fields_ = [
  ('lBand', ctypes.c_int), #[0...............n] more bands means more memory & cpu usage
  ('fBandwidth', ctypes.c_float), #[0.1...........<10] in octaves - fQ is not in use (Bandwidth has a priority over fQ)
  ('fQ', ctypes.c_float), #[0...............1] the EE kinda definition (linear) (if Bandwidth is not in use)
  ('fCenter', ctypes.c_float), #[1Hz..<info.freq/2] in Hz
  ('fGain', ctypes.c_float), #[-15dB...0...+15dB] in dB
  ('lChannel', ctypes.c_float), #BASS_BFX_CHANxxx flag/s
 ]

#Reverb
class BASS_BFX_REVERB(ctypes.Structure):
 _fields_ = [
  ('fLevel', ctypes.c_float), #[0....1....n] linear
  ('lDelay', ctypes.c_int), #[1200..10000]
 ]

#Low Pass Filter
class BASS_BFX_LPF(ctypes.Structure):
 _fields_ = [
  ('fResonance', ctypes.c_float), #[0.01...........10]
  ('fCutOffFreq', ctypes.c_float), #[1Hz...info.freq/2] cutoff frequency
  ('lChannel', ctypes.c_int), #BASS_BFX_CHANxxx flag/s
 ]

#Swap, remap and mix
class BASS_BFX_MIX(ctypes.Structure):
 _fields_ = [
  ('lChannel', ctypes.c_int), #an array of channels to mix using BASS_BFX_CHANxxx flag/s (lChannel[0] is left channel...)
 ]

#Dynamic Amplification
class BASS_BFX_DAMP(ctypes.Structure):
 _fields_ = [
  ('fTarget', ctypes.c_float), #target volume level						[0<......1] linear
  ('fQuiet', ctypes.c_float), #quiet  volume level						[0.......1] linear
  ('fRate', ctypes.c_float), #amp adjustment rate						[0.......1] linear
  ('fGain', ctypes.c_float), #amplification level						[0...1...n] linear
  ('fDelay', ctypes.c_float), #delay in seconds before increasing level	[0.......n] linear
  ('lChannel', ctypes.c_int), #BASS_BFX_CHANxxx flag/s
 ]

#Auto WAH
class BASS_BFX_AUTOWAH(ctypes.Structure):
 _fields_ = [
  ('fDryMix', ctypes.c_float), #dry (unaffected) signal mix				[-2......2]
  ('fWetMix', ctypes.c_float), #wet (affected) signal mix				[-2......2]
  ('fFeedback', ctypes.c_float), #feedback									[-1......1]
  ('fRate', ctypes.c_float), #rate of sweep in cycles per second		[0<....<10]
  ('fRange', ctypes.c_float), #sweep range in octaves					[0<....<10]
  ('fFreq', ctypes.c_float), #base frequency of sweep Hz				[0<...1000]
  ('lChannel', ctypes.c_int), #BASS_BFX_CHANxxx flag/s
 ]

#Echo 2
class BASS_BFX_ECHO2(ctypes.Structure):
 _fields_ = [
  ('fDryMix', ctypes.c_float), #dry (unaffected) signal mix				[-2......2]
  ('fWetMix', ctypes.c_float), #wet (affected) signal mix				[-2......2]
  ('fFeedback', ctypes.c_float), #feedback									[-1......1]
  ('fDelay', ctypes.c_float), #delay sec								[0<......n]
  ('lChannel', ctypes.c_int), #BASS_BFX_CHANxxx flag/s
 ]

#Phaser
class BASS_BFX_PHASER(ctypes.Structure):
 _fields_ = [
  ('fDryMix', ctypes.c_float), #dry (unaffected) signal mix				[-2......2]
  ('fWetMix', ctypes.c_float), #wet (affected) signal mix				[-2......2]
  ('fFeedback', ctypes.c_float), #feedback									[-1......1]
  ('fRate', ctypes.c_float), #rate of sweep in cycles per second		[0<....<10]
  ('fRange', ctypes.c_float), #sweep range in octaves					[0<....<10]
  ('fFreq', ctypes.c_float), #base frequency of sweep					[0<...1000]
  ('lChannel', ctypes.c_int), #BASS_BFX_CHANxxx flag/s
 ]

#Echo 3
class BASS_BFX_ECHO3(ctypes.Structure):
 _fields_ = [
  ('fDryMix', ctypes.c_float), #dry (unaffected) signal mix				[-2......2]
  ('fWetMix', ctypes.c_float), #wet (affected) signal mix				[-2......2]
  ('fDelay', ctypes.c_float), #delay sec								[0<......n]
  ('lChannel', ctypes.c_int), #BASS_BFX_CHANxxx flag/s
 ]

#Chorus
class BASS_BFX_CHORUS(ctypes.Structure):
 _fields_ = [
  ('fDryMix', ctypes.c_float), #dry (unaffected) signal mix				[-2......2]
  ('fWetMix', ctypes.c_float), #wet (affected) signal mix				[-2......2]
  ('fFeedback', ctypes.c_float), #feedback									[-1......1]
  ('fMinSweep', ctypes.c_float), #minimal delay ms							[0<...6000]
  ('fMaxSweep', ctypes.c_float), #maximum delay ms							[0<...6000]
  ('fRate', ctypes.c_float), #rate ms/s								[0<...1000]
  ('lChannel', ctypes.c_int), #BASS_BFX_CHANxxx flag/s
 ]

#All Pass Filter
class BASS_BFX_APF(ctypes.Structure):
 _fields_ = [
  ('fGain', ctypes.c_float), #reverberation time						[-1=<..<=1]
  ('fDelay', ctypes.c_float), #delay sec								[0<....<=n]
  ('lChannel', ctypes.c_int), #BASS_BFX_CHANxxx flag/s
 ]

#Compressor
class BASS_BFX_COMPRESSOR(ctypes.Structure):
 _fields_ = [
  ('fThreshold', ctypes.c_float), #compressor threshold						[0<=...<=1]
  ('fAttacktime', ctypes.c_float), #attack time ms							[0<.<=1000]
  ('fReleasetime', ctypes.c_float), #release time ms							[0<.<=5000]
  ('lChannel', ctypes.c_int), #BASS_BFX_CHANxxx flag/s
 ]

#Distortion
class BASS_BFX_DISTORTION(ctypes.Structure):
 _fields_ = [
  ('fDrive', ctypes.c_float), #distortion drive							[0<=...<=5]
  ('fDryMix', ctypes.c_float), #dry (unaffected) signal mix				[-5<=..<=5]
  ('fWetMix', ctypes.c_float), #wet (affected) signal mix				[-5<=..<=5]
  ('fFeedback', ctypes.c_float), #feedback									[-1<=..<=1]
  ('fVolume', ctypes.c_float), #distortion volume						[0=<...<=2]
  ('lChannel', ctypes.c_int), #BASS_BFX_CHANxxx flag/s
 ]

#Compressor 2
class BASS_BFX_COMPRESSOR2(ctypes.Structure):
 _fields_ = [
  ('fGain', ctypes.c_float), #output gain of signal after compression	[-60....60] in dB
  ('fThreshold', ctypes.c_float), #point at which compression begins		[-60.....0] in dB
  ('fRatio', ctypes.c_float), #compression ratio						[1.......n]
  ('fAttack', ctypes.c_float), #attack time in ms						[0.01.1000]
  ('fRelease', ctypes.c_float), #release time in ms						[0.01.5000]
  ('lChannel', ctypes.c_int), #BASS_BFX_CHANxxx flag/s
 ]

class BASS_BFX_ENV_NODE(ctypes.Structure):
 _fields_ = [
  ('pos', ctypes.c_long), #node position in seconds (1st envelope node must be at position 0)
  ('val', ctypes.c_float), #node value
 ]

#Volume envelope
class BASS_BFX_VOLUME_ENV(ctypes.Structure):
 _fields_ = [
  ('lChannel', ctypes.c_int), #BASS_BFX_CHANxxx flag/s
  ('lNodeCount', ctypes.c_int), #number of nodes
  ('pNodes', BASS_BFX_ENV_NODE), #the nodes
  ('bFollow', ctypes.c_bool), #follow source position
 ]

#BiQuad Filters

class BASS_BFX_BQF(ctypes.Structure):
 _fields_ = [
  ('lFilter', ctypes.c_int), #BASS_BFX_BQF_xxx filter types
  ('fCenter', ctypes.c_float), #[1Hz..<info.freq/2] Cutoff (central) frequency in Hz
  ('fGain', ctypes.c_float), #[-15dB...0...+15dB] Used only for PEAKINGEQ and Shelving filters in dB
  ('fBandwidth', ctypes.c_float), #[0.1...........<10] Bandwidth in octaves (fQ is not in use (fBandwidth has a priority over fQ))
   #(between -3 dB frequencies for BANDPASS and NOTCH or between midpoint
   #(fGgain/2) gain frequencies for PEAKINGEQ)
  ('fQ', ctypes.c_float), #[0.1.............1] The EE kinda definition (linear) (if fBandwidth is not in use)
  ('fS', ctypes.c_float), #[0.1.............1] A "shelf slope" parameter (linear) (used only with Shelving filters)
   #when fS = 1, the shelf slope is as steep as you can get it and remain monotonically
   #increasing or decreasing gain with frequency.
  ('lChannel', ctypes.c_int), #BASS_BFX_CHANxxx flag/s
 ]

"""TEMPO / PITCH SCALING / SAMPLERATE"""

"""NOTES: 1. Supported only - mono / stereo - channels
2. Enable Tempo supported flags in BASS_FX_TempoCreate and the others to source handle.
tempo attributes (BASS_ChannelSet/GetAttribute)"""

(
 BASS_ATTRIB_TEMPO,
 BASS_ATTRIB_TEMPO_PITCH,
 BASS_ATTRIB_TEMPO_FREQ,
) = range(0x10000, 0x10000+3)

#tempo attributes options
#[option]											[value]

(
 BASS_ATTRIB_TEMPO_OPTION_USE_AA_FILTER, #TRUE (default) / FALSE
 BASS_ATTRIB_TEMPO_OPTION_AA_FILTER_LENGTH, #32 default (8 .. 128 taps)
 BASS_ATTRIB_TEMPO_OPTION_USE_QUICKALGO, #TRUE / FALSE (default)
 BASS_ATTRIB_TEMPO_OPTION_SEQUENCE_MS, #82 default, 0 = automatic
 BASS_ATTRIB_TEMPO_OPTION_SEEKWINDOW_MS, #28 default, 0 = automatic
 BASS_ATTRIB_TEMPO_OPTION_OVERLAP_MS, #8  default
 BASS_ATTRIB_TEMPO_OPTION_PREVENT_CLICK, #TRUE / FALSE (default)
) = range(0x10000, 0x10000+7)

#HSTREAM BASS_FXDEF(BASS_FX_TempoCreate)(DWORD chan, DWORD flags);
BASS_FX_TempoCreate = func_type(pybass.HSTREAM, ctypes.c_ulong, ctypes.c_ulong)(('BASS_FX_TempoCreate', bass_fx_module))

#DWORD BASS_FXDEF(BASS_FX_TempoGetSource)(HSTREAM chan);
BASS_FX_TempoGetSource = func_type(ctypes.c_ulong, pybass.HSTREAM)(('BASS_FX_TempoGetSource', bass_fx_module))

#float BASS_FXDEF(BASS_FX_TempoGetRateRatio)(HSTREAM chan);
BASS_FX_TempoGetRateRatio = func_type(ctypes.c_float, pybass.HSTREAM)(('BASS_FX_TempoGetRateRatio', bass_fx_module))

"""R E V E R S E"""
"""NOTES: 1. MODs won't load without BASS_MUSIC_PRESCAN flag.
2. Enable Reverse supported flags in BASS_FX_ReverseCreate and the others to source handle."""

#reverse attribute (BASS_ChannelSet/GetAttribute)
BASS_ATTRIB_REVERSE_DIR = 0x11000

#playback directions
BASS_FX_RVS_REVERSE = -1
BASS_FX_RVS_FORWARD = 1

#HSTREAM BASS_FXDEF(BASS_FX_ReverseCreate)(DWORD chan, float dec_block, DWORD flags);
BASS_FX_ReverseCreate = func_type(pybass.HSTREAM, ctypes.c_ulong, ctypes.c_float, ctypes.c_ulong)(('BASS_FX_ReverseCreate', bass_fx_module))

#DWORD BASS_FXDEF(BASS_FX_ReverseGetSource)(HSTREAM chan);
BASS_FX_ReverseGetSource = func_type(ctypes.c_ulong, pybass.HSTREAM)(('BASS_FX_ReverseGetSource', bass_fx_module))

"""B P M (Beats Per Minute)"""

#bpm flags
BASS_FX_BPM_BKGRND = 1 #if in use, then you can do other processing while detection's in progress. Not available in MacOSX yet. (BPM/Beat)
BASS_FX_BPM_MULT2 = 2 #if in use, then will auto multiply bpm by 2 (if BPM < minBPM*2)

#translation options
(
 BASS_FX_BPM_TRAN_X2, #multiply the original BPM value by 2 (may be called only once & will change the original BPM as well!)
 BASS_FX_BPM_TRAN_2FREQ, #BPM value to Frequency
 BASS_FX_BPM_TRAN_FREQ2, #Frequency to BPM value
 BASS_FX_BPM_TRAN_2PERCENT, #BPM value to Percents
 BASS_FX_BPM_TRAN_PERCENT2	, #Percents to BPM value
) = range(5)

#typedef void (CALLBACK BPMPROCESSPROC)(DWORD chan, float percent);
BPMPROCESSPROC = func_type(None, ctypes.c_float)

#typedef void (CALLBACK BPMPROC)(DWORD chan, float bpm, void *user);
BPMPROC = func_type(None, ctypes.c_long, ctypes.c_float, ctypes.c_void_p)

#float BASS_FXDEF(BASS_FX_BPM_DecodeGet)(DWORD chan, double startSec, double endSec, DWORD minMaxBPM, DWORD flags, BPMPROCESSPROC *proc);
BASS_FX_BPM_DecodeGet = func_type(ctypes.c_float, ctypes.c_ulong, ctypes.c_double, ctypes.c_double, ctypes.c_ulong, ctypes.c_ulong, BPMPROCESSPROC)(('BASS_FX_BPM_DecodeGet', bass_fx_module))

#BOOL BASS_FXDEF(BASS_FX_BPM_CallbackSet)(DWORD handle, BPMPROC *proc, double period, DWORD minMaxBPM, DWORD flags, void *user);
BASS_FX_BPM_CallbackSet = func_type(ctypes.c_bool, ctypes.c_ulong, BPMPROC, ctypes.c_double, ctypes.c_ulong, ctypes.c_ulong, ctypes.c_void_p)(('BASS_FX_BPM_CallbackSet', bass_fx_module))

#BOOL BASS_FXDEF(BASS_FX_BPM_CallbackReset)(DWORD handle);
BASS_FX_BPM_CallbackReset = func_type(ctypes.c_bool, ctypes.c_ulong)(('BASS_FX_BPM_CallbackReset', bass_fx_module))

#float BASS_FXDEF(BASS_FX_BPM_Translate)(DWORD handle, float val2tran, DWORD trans);
BASS_FX_BPM_Translate = func_type(ctypes.c_float, ctypes.c_ulong, ctypes.c_float, ctypes.c_ulong)(('BASS_FX_BPM_Translate', bass_fx_module))

#BOOL BASS_FXDEF(BASS_FX_BPM_Free)(DWORD handle);
BASS_FX_BPM_Free = func_type(ctypes.c_bool, ctypes.c_ulong)(('BASS_FX_BPM_Free', bass_fx_module))

""" Beat """

#typedef void (CALLBACK BPMBEATPROC)(DWORD chan, double beatpos, void *user);
BPMBEATPROC = func_type(None, ctypes.c_ulong, ctypes.c_double, ctypes.c_void_p)

#BOOL BASS_FXDEF(BASS_FX_BPM_BeatCallbackSet)(DWORD handle, BPMBEATPROC *proc, void *user);
BASS_FX_BPM_BeatCallbackSet = func_type(ctypes.c_bool, ctypes.c_ulong, ctypes.c_void_p)(('BASS_FX_BPM_BeatCallbackSet', bass_fx_module))

#BOOL BASS_FXDEF(BASS_FX_BPM_BeatCallbackReset)(DWORD handle);
BASS_FX_BPM_BeatCallbackReset = func_type(ctypes.c_bool, ctypes.c_ulong)(('BASS_FX_BPM_BeatCallbackReset', bass_fx_module))

#BOOL BASS_FXDEF(BASS_FX_BPM_BeatDecodeGet)(DWORD chan, double startSec, double endSec, DWORD flags, BPMBEATPROC *proc, void *user);
BASS_FX_BPM_BeatDecodeGet = func_type(ctypes.c_bool, ctypes.c_ulong, ctypes.c_double, ctypes.c_double, ctypes.c_ulong, BPMBEATPROC, ctypes.c_void_p)(('BASS_FX_BPM_BeatDecodeGet', bass_fx_module))

#BOOL BASS_FXDEF(BASS_FX_BPM_BeatSetParameters)(DWORD handle, float bandwidth, float centerfreq, float beat_rtime);
BASS_FX_BPM_BeatSetParameters = func_type(ctypes.c_bool, ctypes.c_ulong, ctypes.c_float, ctypes.c_float, ctypes.c_float)(('BASS_FX_BPM_BeatSetParameters', bass_fx_module))

#BOOL BASS_FXDEF(BASS_FX_BPM_BeatGetParameters)(DWORD handle, float *bandwidth, float *centerfreq, float *beat_rtime);
BASS_FX_BPM_BeatGetParameters = func_type(ctypes.c_bool, ctypes.c_ulong, ctypes.c_float, ctypes.c_float, ctypes.c_float)(('BASS_FX_BPM_BeatGetParameters', bass_fx_module))

#BOOL BASS_FXDEF(BASS_FX_BPM_BeatFree)(DWORD handle);
BASS_FX_BPM_BeatFree = func_type(ctypes.c_bool, ctypes.c_ulong)(('BASS_FX_BPM_BeatFree', bass_fx_module))



from __future__ import absolute_import
import platform
from . import pybassopus
if platform.system() == 'Windows':
 from . import pybasswma
if platform.system() != 'Darwin':
 from . import pybass_aac
 from . import pybass_alac
 from . import pybassflac
 from . import pybassmidi



from __future__ import absolute_import
from ctypes import pointer
from functools import partial
from .main import bass_call, update_3d_system
from .external.pybass import *


def _getter(base_prop, attr, obj):
    """

    Args:
      base_prop: 
      attr: 
      obj: 

    Returns:

    """
    return getattr(getattr(obj, base_prop), attr)


def _setter(base_prop, subattr, obj, val):
    """

    Args:
      base_prop: 
      subattr: 
      obj: 
      val: 

    Returns:

    """
    old = getattr(obj, base_prop)
    setattr(old, subattr, val)
    setattr(obj, base_prop, old)


class Listener(object):
    """ """
    def get_3d_position(self):
        """ """
        res = {
            "position": BASS_3DVECTOR(),
            "velocity": BASS_3DVECTOR(),
            "front": BASS_3DVECTOR(),
            "top": BASS_3DVECTOR(),
        }
        bass_call(
            BASS_Get3DPosition,
            pointer(res["position"]),
            pointer(res["velocity"]),
            pointer(res["front"]),
            pointer(res["top"]),
        )
        return res

    @update_3d_system
    def set_3d_position(self, position=None, velocity=None, front=None, top=None):
        """Sets the position, velocity, and orientation of the listener (ie. the player).

        Args:
          position:  (Default value = None)
          velocity:  (Default value = None)
          front:  (Default value = None)
          top:  (Default value = None)

        Returns:

        """
        old = self.get_3d_position()
        if position is None:
            position = old["position"]
        if velocity is None:
            velocity = old["velocity"]
        if front is None:
            front = old["front"]
        if top is None:
            top = old["top"]
        position = pointer(position)
        velocity = pointer(velocity)
        front = pointer(front)
        top = pointer(top)
        bass_call(BASS_Set3DPosition, position, velocity, front, top)

    def get_position(self):
        """ """
        return self.get_3d_position()["position"]

    def set_position(self, position):
        """

        Args:
          position: 

        Returns:

        """
        self.set_3d_position(position=position)

    position = property(fget=get_position, fset=set_position)

    x = property(
        fget=partial(_getter, "position", "x"), fset=partial(_setter, "position", "x")
    )

    y = property(
        fget=partial(_getter, "position", "y"), fset=partial(_setter, "position", "y")
    )

    z = property(
        fget=partial(_getter, "position", "z"), fset=partial(_setter, "position", "z")
    )

    def get_velocity(self):
        """ """
        return self.get_3d_position()["velocity"]

    def set_velocity(self, velocity):
        """

        Args:
          velocity: 

        Returns:

        """
        self.set_3d_position(velocity=velocity)

    velocity = property(fget=get_velocity, fset=set_velocity)

    x_velocity = property(
        fget=partial(_getter, "velocity", "x"), fset=partial(_setter, "velocity", "x")
    )

    y_velocity = property(
        fget=partial(_getter, "velocity", "y"), fset=partial(_setter, "velocity", "y")
    )

    z_velocity = property(
        fget=partial(_getter, "velocity", "z"), fset=partial(_setter, "velocity", "z")
    )

    def get_front(self):
        """ """
        return self.get_3d_position()["front"]

    def set_front(self, front):
        """

        Args:
          front: 

        Returns:

        """
        self.set_3d_position(front=front)

    front = property(fget=get_front, fset=set_front)

    front_x = property(
        fget=partial(_getter, "front", "x"), fset=partial(_setter, "front", "x")
    )

    front_y = property(
        fget=partial(_getter, "front", "y"), fset=partial(_setter, "front", "y")
    )

    front_z = property(
        fget=partial(_getter, "front", "z"), fset=partial(_setter, "front", "z")
    )

    def get_top(self):
        """ """
        return self.get_3d_position()["top"]

    def set_top(self, top):
        """

        Args:
          top: 

        Returns:

        """
        self.set_3d_position(top=top)

    top = property(fget=get_top, fset=set_top)

    top_x = property(
        fget=partial(_getter, "front", "x"), fset=partial(_setter, "front", "x")
    )

    top_y = property(
        fget=partial(_getter, "front", "y"), fset=partial(_setter, "front", "y")
    )

    top_z = property(
        fget=partial(_getter, "front", "z"), fset=partial(_setter, "front", "z")
    )



# Copyright(c) Andrew Evans 2013
# BSD license

__version__ = '0.1'
__versionTime__ = '2013-03-28'
__author__ = 'Andrew Evans <themindflows@gmail.com>'
__doc__ = '''
pybasswasapi.py - is ctypes python module for WASAPI (http://www.un4seen.com).

BASSWASAPI is basically a wrapper for WASAPI drivers
BASSWASAPI requires a soundcard with WASAPI drivers. 
'''

import sys, ctypes, platform

from . import pybass
from .paths import x86_path, x64_path
import libloader


HSTREAM = pybass.HSTREAM
BASS_FILEPROCS = pybass.BASS_FILEPROCS


basswasapi_module = libloader.load_library('basswasapi', x86_path=x86_path, x64_path=x64_path)
func_type = libloader.get_functype()

# Additional error codes returned by BASS_ErrorGetCode
BASS_ERROR_WASAPI = 5000

# Device info structure
class BASS_WASAPI_DEVICEINFO(ctypes.Structure):
    _fields_ = [
        ('name', ctypes.c_char_p),
        ('id', ctypes.c_char_p),
        ('type', ctypes.c_ulong),
        ('flags', ctypes.c_ulong),
        ('minperiod', ctypes.c_float),
        ('defperiod', ctypes.c_float),
        ('mixfreq', ctypes.c_ulong),
        ('mixchans', ctypes.c_ulong)      
    ]

class BASS_WASAPI_INFO(ctypes.Structure):
    _fields_ = [
        ('initflag', ctypes.c_ulong),
        ('freq', ctypes.c_ulong),
        ('chans', ctypes.c_ulong),
        ('format', ctypes.c_ulong),
        ('buflen', ctypes.c_ulong),
        ('volmax', ctypes.c_float),
        ('volmin', ctypes.c_float),
        ('volstep', ctypes.c_float)
    ]

# BASS_WASAPI_DEVICEINFO "type"
BASS_WASAPI_TYPE_NETWORKDEVICE = 0
BASS_WASAPI_TYPE_SPEAKERS = 1
BASS_WASAPI_TYPE_LINELEVEL = 2
BASS_WASAPI_TYPE_HEADPHONES = 3
BASS_WASAPI_TYPE_MICROPHONE = 4
BASS_WASAPI_TYPE_HEADSET = 5
BASS_WASAPI_TYPE_HANDSET = 6
BASS_WASAPI_TYPE_DIGITAL = 7
BASS_WASAPI_TYPE_SPDIF = 8
BASS_WASAPI_TYPE_HDMI = 9
BASS_WASAPI_TYPE_UNKNOWN = 10

# BASS_WASAPI_DEVICEINFO flags
BASS_DEVICE_ENABLED = 1
BASS_DEVICE_DEFAULT = 2
BASS_DEVICE_INIT = 4
BASS_DEVICE_LOOPBACK = 8
BASS_DEVICE_INPUT = 16
BASS_DEVICE_UNPLUGGED = 32
BASS_DEVICE_DISABLED = 64

# BASS_WASAPI_Init flags
BASS_WASAPI_EXCLUSIVE = 1
BASS_WASAPI_AUTOFORMAT = 2
BASS_WASAPI_BUFFER = 4
BASS_WASAPI_SESSIONVOL = 8
BASS_WASAPI_EVENT = 16

# BASS_WASAPI_INFO "format"
BASS_WASAPI_FORMAT_FLOAT = 0
BASS_WASAPI_FORMAT_8BIT = 1
BASS_WASAPI_FORMAT_16BIT = 2
BASS_WASAPI_FORMAT_24BIT = 3
BASS_WASAPI_FORMAT_32BIT = 4

# BASS_WASAPI_Set/GetVolume "curve"
BASS_WASAPI_CURVE_DB = 0
BASS_WASAPI_CURVE_LINEAR = 1
BASS_WASAPI_CURVE_WINDOWS = 2

#typedef DWORD (CALLBACK WASAPIPROC)(void *buffer, DWORD length, void *user);
WASAPIPROC = func_type(ctypes.c_ulong, ctypes.c_void_p, ctypes.c_ulong, ctypes.c_void_p)

#typedef void (CALLBACK WASAPINOTIFYPROC)(DWORD notify, DWORD device, void *user);
WASAPINOTIFYPROC = func_type(ctypes.c_ulong, ctypes.c_ulong, ctypes.c_void_p)

# Device notifications
BASS_WASAPI_NOTIFY_ENABLED = 0
BASS_WASAPI_NOTIFY_DISABLED = 1
BASS_WASAPI_NOTIFY_DEFOUTPUT = 2
BASS_WASAPI_NOTIFY_DEFINPUT = 3


# DWORD BASSWASAPIDEF(BASS_WASAPI_GetVersion)();
BASS_WASAPI_GetVersion = func_type(HSTREAM)(('BASS_WASAPI_GetVersion', basswasapi_module))
# BOOL BASSWASAPIDEF(BASS_WASAPI_SetNotify)(WASAPINOTIFYPROC *proc, void *user);
BASS_WASAPI_SetNotify = func_type(HSTREAM, ctypes.POINTER(WASAPINOTIFYPROC), ctypes.c_void_p)(('BASS_WASAPI_SetNotify', basswasapi_module))
# BOOL BASSWASAPIDEF(BASS_WASAPI_GetDeviceInfo)(DWORD device, BASS_WASAPI_DEVICEINFO *info);
BASS_WASAPI_GetDeviceInfo = func_type(HSTREAM, ctypes.c_ulong, ctypes.POINTER(BASS_WASAPI_DEVICEINFO))(('BASS_WASAPI_GetDeviceInfo', basswasapi_module))
# float BASSDEF(BASS_WASAPI_GetDeviceLevel)(DWORD device, int chan);
BASS_WASAPI_GetDeviceLevel = func_type(HSTREAM, ctypes.c_ulong, ctypes.c_int)(('BASS_WASAPI_GetDeviceLevel', basswasapi_module))
# BOOL BASSWASAPIDEF(BASS_WASAPI_SetDevice)(DWORD device);
BASS_WASAPI_SetDevice = func_type(HSTREAM, ctypes.c_ulong)(('BASS_WASAPI_SetDevice', basswasapi_module))
# DWORD BASSWASAPIDEF(BASS_WASAPI_GetDevice)();
BASS_WASAPI_GetDevice = func_type(HSTREAM)(('BASS_WASAPI_GetDevice', basswasapi_module))
# DWORD BASSWASAPIDEF(BASS_WASAPI_CheckFormat)(DWORD device, DWORD freq, DWORD chans, DWORD flags);
BASS_WASAPI_CheckFormat = func_type(HSTREAM, ctypes.c_ulong, ctypes.c_ulong, ctypes.c_ulong, ctypes.c_ulong)(('BASS_WASAPI_CheckFormat', basswasapi_module))
# BOOL BASSWASAPIDEF(BASS_WASAPI_Init)(int device, DWORD freq, DWORD chans, DWORD flags, float buffer, float period, WASAPIPROC *proc, void *user);
BASS_WASAPI_Init = func_type(HSTREAM, ctypes.c_int, ctypes.c_ulong, ctypes.c_ulong, ctypes.c_ulong, ctypes.c_float, ctypes.c_float, WASAPIPROC, ctypes.c_void_p)(('BASS_WASAPI_Init', basswasapi_module))
# BOOL BASSWASAPIDEF(BASS_WASAPI_Free)();
BASS_WASAPI_Free = func_type(HSTREAM)(('BASS_WASAPI_Free', basswasapi_module))
# BOOL BASSWASAPIDEF(BASS_WASAPI_GetInfo)(BASS_WASAPI_INFO *info);
BASS_WASAPI_GetInfo = func_type(HSTREAM, ctypes.POINTER(BASS_WASAPI_INFO))(('BASS_WASAPI_GetInfo', basswasapi_module))
# float BASSWASAPIDEF(BASS_WASAPI_GetCPU)();
BASS_WASAPI_GetCPU = func_type(HSTREAM)(('BASS_WASAPI_GetCPU', basswasapi_module))
# BOOL BASSWASAPIDEF(BASS_WASAPI_Lock)(BOOL lock);
BASS_WASAPI_Lock = func_type(HSTREAM, ctypes.c_bool)(('BASS_WASAPI_Lock', basswasapi_module))
# BOOL BASSWASAPIDEF(BASS_WASAPI_Start)();
BASS_WASAPI_Start = func_type(HSTREAM)(('BASS_WASAPI_Start', basswasapi_module))
# BOOL BASSWASAPIDEF(BASS_WASAPI_Stop)(BOOL reset);
BASS_WASAPI_Stop = func_type(ctypes.c_bool, ctypes.c_bool)(('BASS_WASAPI_Stop', basswasapi_module))
# BOOL BASSWASAPIDEF(BASS_WASAPI_IsStarted)();
BASS_WASAPI_IsStarted = func_type(HSTREAM)(('BASS_WASAPI_IsStarted', basswasapi_module))
# BOOL BASSWASAPIDEF(BASS_WASAPI_SetVolume)(DWORD curve, float volume);
BASS_WASAPI_SetVolume = func_type(HSTREAM, ctypes.c_ulong, ctypes.c_float)(('BASS_WASAPI_SetVolume', basswasapi_module))
# float BASSWASAPIDEF(BASS_WASAPI_GetVolume)(DWORD curve);
BASS_WASAPI_GetVolume = func_type(HSTREAM, ctypes.c_ulong)(('BASS_WASAPI_GetVolume', basswasapi_module))
# BOOL BASSWASAPIDEF(BASS_WASAPI_SetMute)(BOOL mute);
BASS_WASAPI_SetMute = func_type(HSTREAM, ctypes.c_bool)(('BASS_WASAPI_SetMute', basswasapi_module))
# BOOL BASSWASAPIDEF(BASS_WASAPI_GetMute)();
BASS_WASAPI_GetMute = func_type(HSTREAM)(('BASS_WASAPI_GetMute', basswasapi_module))
# DWORD BASSWASAPIDEF(BASS_WASAPI_PutData)(void *buffer, DWORD length);
BASS_WASAPI_PutData = func_type(HSTREAM, ctypes.c_void_p, ctypes.c_ulong)(('BASS_WASAPI_PutData', basswasapi_module))
# DWORD BASSDEF(BASS_WASAPI_GetData)(void *buffer, DWORD length);
BASS_WASAPI_GetData = func_type(HSTREAM, ctypes.c_void_p, ctypes.c_ulong)(('BASS_WASAPI_GetData', basswasapi_module))
# DWORD BASSDEF(BASS_WASAPI_GetLevel)();
BASS_WASAPI_GetLevel = func_type(HSTREAM)(('BASS_WASAPI_GetLevel', basswasapi_module))




from __future__ import absolute_import
from functools import partial
import platform
from ctypes import c_char_p, c_float, pointer, string_at
from .external.pybass import *
from . import config
from .main import bass_call, bass_call_0, EAX_ENVIRONMENTS, update_3d_system

_getter = lambda func, key, obj: func(obj)[key]
_setter = lambda func, kwarg, obj, val: func(obj, **{kwarg: val})


class Output(object):
    """ """
    def __init__(self, device=-1, frequency=44100, flags=0, window=0, clsid=None):
        try:
            self.use_default_device()
        except:
            pass
        self._device = device
        self.frequency = frequency
        self.flags = flags
        self.window = window
        self.clsid = clsid
        self.init_device(
            device=device, frequency=frequency, flags=flags, window=window, clsid=clsid
        )
        self.config = config.BassConfig()
        self.proxy = None

    def init_device(
        self, device=None, frequency=None, flags=None, window=None, clsid=None
    ):
        """

        Args:
          device:  (Default value = None)
          frequency:  (Default value = None)
          flags:  (Default value = None)
          window:  (Default value = None)
          clsid:  (Default value = None)

        Returns:

        """
        if device is None:
            device = self._device
        self._device = device
        if frequency is None:
            frequency = self.frequency
        self.frequency = frequency
        if flags is None:
            flags = self.flags
        self.flags = flags
        if window is None:
            window = self.window
        self.window = window
        if clsid is None:
            clsid = self.clsid
        self.clsid = clsid
        if (
            platform.system() == "Linux" and device == -1
        ):  # Bass wants default device set to 1 on linux
            device = 1
        bass_call(BASS_Init, device, frequency, flags, window, clsid)

    def start(self):
        """ """
        return bass_call(BASS_Start)

    def pause(self):
        """ """
        return bass_call(BASS_Pause)

    def stop(self):
        """ """
        return bass_call(BASS_Stop)

    def get_device(self):
        """ """
        return bass_call_0(BASS_GetDevice)

    def set_device(self, device):
        """

        Args:
          device: 

        Returns:

        """
        if device == self._device:
            return
        self.free()
        self.init_device(device=device)
        return bass_call(BASS_SetDevice, device)

    device = property(fget=get_device, fset=set_device)

    def get_volume(self):
        """ """
        volume = BASS_GetConfig(BASS_CONFIG_GVOL_STREAM)
        if volume:
            volume = volume / 100.0
        return volume

    def set_volume(self, volume):
        """

        Args:
          volume: 

        Returns:

        """
        # Pass in a float 0.0 to 100.0 and watch the volume magically change
        return bass_call(
            BASS_SetConfig, BASS_CONFIG_GVOL_STREAM, int(round(volume * 100, 2))
        )

    volume = property(get_volume, set_volume)

    @staticmethod
    def free():
        """ """
        return bass_call(BASS_Free)

    def get_proxy(self):
        """ """
        ptr = bass_call(BASS_GetConfigPtr, BASS_CONFIG_NET_PROXY)
        return string_at(ptr)

    def set_proxy(self, proxy):
        """

        Args:
          proxy: 

        Returns:

        """
        self.proxy = c_char_p(proxy)
        return bass_call(BASS_SetConfigPtr, BASS_CONFIG_NET_PROXY, self.proxy)

    def use_default_device(self, use=True):
        """

        Args:
          use:  (Default value = True)

        Returns:

        """
        return bass_call(BASS_SetConfig, BASS_CONFIG_DEV_DEFAULT, use)

    @staticmethod
    def get_device_names():
        """Convenience method that returns a list of device names that are considered
        	valid by bass.

        Args:

        Returns:
          

        """
        result = []  # empty list to start.
        info = BASS_DEVICEINFO()
        count = 1
        while BASS_GetDeviceInfo(count, ctypes.byref(info)):
            if info.flags & BASS_DEVICE_ENABLED:
                retrieved = info.name
                if platform.system() == "Windows":
                    retrieved = retrieved.decode("mbcs")
                elif platform.system() == "Darwin":
                    retrieved = retrieved.decode("utf-8")
                retrieved = retrieved.replace("(", "").replace(")", "").strip()
                result.append(retrieved)
            count += 1
        return result

    def find_device_by_name(self, name):
        """

        Args:
          name: 

        Returns:

        """
        return self.get_device_names().index(name) + 1

    def find_default_device(self):
        """ """
        try:
            return self.get_device_names().index("Default") + 1
        except:
            return -1

    def find_user_provided_device(self, device_name):
        """

        Args:
          device_name: 

        Returns:

        """
        try:
            return self.find_device_by_name(device_name)
        except ValueError:
            return self.find_default_device()


class ThreeDOutput(Output):
    """ """
    def __init__(self, flags=BASS_DEVICE_3D, *args, **kwargs):
        super(ThreeDOutput, self).__init__(flags=flags, *args, **kwargs)

    def get_3d_factors(self):
        """ """
        res = {
            "distance_factor": c_float(),
            "rolloff": c_float(),
            "doppler_factor": c_float(),
        }
        bass_call(
            BASS_Get3DFactors,
            pointer(res["distance_factor"]),
            pointer(res["rolloff"]),
            pointer(res["doppler_factor"]),
        )
        return {k: res[k].value for k in res}

    @update_3d_system
    def set_3d_factors(self, distance_factor=-1, rolloff=-1, doppler_factor=-1):
        """

        Args:
          distance_factor:  (Default value = -1)
          rolloff:  (Default value = -1)
          doppler_factor:  (Default value = -1)

        Returns:

        """
        conversions = {"meters": 1.0, "yards": 0.9144, "feet": 0.3048}
        if distance_factor in conversions:
            distance_factor = conversions[distance_factor]
        return bass_call(BASS_Set3DFactors, distance_factor, rolloff, doppler_factor)

    distance_factor = property(
        fget=partial(_getter, get_3d_factors, "distance_factor"),
        fset=partial(_setter, set_3d_factors, "distance_factor"),
    )

    rolloff = property(
        fget=partial(_getter, get_3d_factors, "rolloff"),
        fset=partial(_setter, set_3d_factors, "rolloff"),
    )

    doppler_factor = property(
        fget=partial(_getter, get_3d_factors, "doppler_factor"),
        fset=partial(_setter, set_3d_factors, "doppler_factor"),
    )

    def set_eax_parameters(self, environment=None, volume=None, decay=None, damp=None):
        """

        Args:
          environment:  (Default value = None)
          volume:  (Default value = None)
          decay:  (Default value = None)
          damp:  (Default value = None)

        Returns:

        """
        def convert_arg(arg):
            """

            Args:
              arg: 

            Returns:

            """
            if arg is None:
                arg = -1
            return arg

        environment = convert_arg(environment)
        if isinstance(environment, basestring) and environment in EAX_ENVIRONMENTS:
            environment = EAX_ENVIRONMENTS[environment]
        volume = convert_arg(volume)
        decay = convert_arg(decay)
        damp = convert_arg(damp)
        bass_call(BASS_SetEAXParameters, environment, volume, decay, damp)

    def get_3d_algorithm(self):
        """ """
        return BASS_GetConfig(BASS_CONFIG_3DALGORITHM)

    def set_3d_algorithm(self, algo):
        """

        Args:
          algo: 

        Returns:

        """
        replacements = {
            "default": BASS_3DALG_DEFAULT,
            "off": BASS_3DALG_OFF,
            "full": BASS_3DALG_FULL,
            "light": BASS_3DALG_LIGHT,
        }
        if algo in replacements:
            algo = replacements[algo]
        return BASS_SetConfig(BASS_CONFIG_3DALGORITHM, algo)



from __future__ import absolute_import
from .external.pybass import *
from .main import bass_call, bass_call_0, BassError, update_3d_system, FlagObject
from ctypes import pointer, c_float, c_long, c_ulong, c_buffer


class Channel(FlagObject):
    """A "channel" can be a sample playback channel (HCHANNEL), a sample stream (HSTREAM), a MOD music (HMUSIC), or a recording (HRECORD).
    Chances are, if you're playing audio, you're using a channel on one level or another.
    Each "Channel" function can be used with one or more of these channel types.
    """

    attribute_mapping = {}

    def __init__(self, handle):
        self.handle = handle
        self.attribute_mapping = {
            "eaxmix": BASS_ATTRIB_EAXMIX,
            "frequency": BASS_ATTRIB_FREQ,
            "pan": BASS_ATTRIB_PAN,
            "volume": BASS_ATTRIB_VOL,
            "byte": BASS_POS_BYTE,
            "decode": BASS_POS_DECODE,
        }

    def add_attributes_to_mapping(self, **attrs):
        self.attribute_mapping.update(**attrs)

    def play(self, restart=False):
        """Starts (or resumes) playback of a sample, stream, MOD music, or recording.

        Args:
          restart (bool):  Specifies whether playback position should be thrown to the beginning of the stream. Defaults to False

        Returns:
            bool: True on success, False otherwise.
        """
        return bass_call(BASS_ChannelPlay, self.handle, restart)

    def play_blocking(self, restart=False):
        """Starts (or resumes) playback, waiting to return until reaching the end of the stream

        Args:
          restart (bool):  Specifies whether playback position should be thrown to the beginning of the stream. Defaults to False.
        """
        self.play(restart=restart)
        while self.is_playing:
            pass

    def pause(self):
        """Pauses a sample, stream, MOD music, or recording.

        Returns:
            bool: True on success, False otherwise.

        raises:
            sound_lib.main.BassError: If this channel isn't currently playing, already paused, or is a decoding channel and thus not playable.
        """
        return bass_call(BASS_ChannelPause, self.handle)

    def is_active(self):
        """Checks if a sample, stream, or MOD music is active (playing) or stalled. Can also check if a recording is in progress.
        """
        return bass_call_0(BASS_ChannelIsActive, self.handle)

    @property
    def is_playing(self):
        """Checks whether the stream is currently playing or recording.

        Returns:
            bool: True if playing, False otherwise.
        """
        return self.is_active() == BASS_ACTIVE_PLAYING

    @property
    def is_paused(self):
        """
        Checks whether the stream is currently paused.

        Returns:
            bool: True if paused, False otherwise.
        """
        return self.is_active() == BASS_ACTIVE_PAUSED

    @property
    def is_stopped(self):
        """
        Checks whether the stream is currently stopped.

        Returns:
            bool: True if stopped, False otherwise.
        """
        return self.is_active() == BASS_ACTIVE_STOPPED

    @property
    def is_stalled(self):
        """
        Checks whether playback of a stream has been stalled.
        This is due to a lack of sample data. Playback will automatically resume once there is sufficient data to do so.

        Returns:
            bool: True if stalled, False otherwise.
        """
        return self.is_active() == BASS_ACTIVE_STALLED

    def get_position(self, mode=BASS_POS_BYTE):
        """Retrieves the playback position of a sample, stream, or MOD music. Can also be used with a recording channel.

        Args:
          mode (str):  How to retrieve the position. Defaults to "byte".

        Returns:
            int: The current position.

        raises:
            sound_lib.main.BassError: If the requested position is not available.
        """
        return bass_call_0(BASS_ChannelGetPosition, self.handle, mode)

    def set_position(self, pos, mode=BASS_POS_BYTE):
        """Sets the playback position of a sample, MOD music, or stream.

        Args:
          pos (int): The position, in units determined by the mode.
          mode:  (str): How to set the position. Defaults to "byte".

        Returns:
            bool: True if the position was set, False otherwise.

        raises:
            sound_lib.main.BassError: If the stream is not a sound_lib.stream.FileStream or the requested position/mode is not available.
        """
        return bass_call(BASS_ChannelSetPosition, self.handle, pos, mode)

    position = property(get_position, set_position)

    def stop(self):
        """Stops a sample, stream, MOD music, or recording."""
        return bass_call(BASS_ChannelStop, self.handle)

    def update(self, length=0):
        """Updates the playback buffer of a stream or MOD music.

        Args:
          length (int): The amount of data to render, in milliseconds...
              0 = default (2 x update period). This is capped at the space available in the buffer.

        Returns:
            bool: True on success, False otherwise.

        raises:
            sound_lib.main.BassError: If this channel has ended or doesn't have an output -buffer.
        """
        return bass_call(BASS_ChannelUpdate, self.handle, length)

    def get_length(self, mode=BASS_POS_BYTE):
        """Retrieves the playback length of this channel.

        Args:
          mode:  How to retrieve the length. Can take either a flag attribute (string) or bass constent (int). Defaults to "byte".

        Returns:
            int: The channel length on success, -1 on failure.

        raises:
            sound_lib.main.BassError: If the requested mode is not available.
        """
        return bass_call_0(BASS_ChannelGetLength, self.handle, mode)

    __len__ = get_length

    def __nonzero__(self):
        return True

    def get_device(self):
        """Retrieves the device in use by this channel.

        returns:
            int: The device number, -1 on failure.
        """
        return bass_call_0(BASS_ChannelGetDevice, self.handle)

    def set_device(self, device):
        """Changes the device in use by this channel. Must be a stream, MOD music or sample.

        Args:
          device: The device to use... 0 = no sound, 1 = first real output device, BASS_NODEVICE = no device.

        Returns:
            bool: True on success, False otherwise.

        raises:
            sound_lib.main.BassError: If device is invalid, device hasn't been initialized, this channel is already using the requested device, the sample format is not supported by the device/drivers or there is insufficient memory.
        """
        bass_call(BASS_ChannelSetDevice, self.handle, device)

    device = property(get_device, set_device)

    def set_fx(self, type, priority=0):
        """Sets an effect on a stream, MOD music, or recording channel.

        Args:
          type: The type of effect
          priority: The priority of the new FX, which determines its position in the DSP chain. DSP/FX with higher priority are applied before those with lower. This parameter has no effect with DX8 effects when the "with FX flag" DX8 effect implementation is used. Defaults to 0.

        Returns:
            A handle to the new effect on success, False otherwise.

        raises:
            sound_lib.main.BassError: If type is invalid, the specified DX8 effect is unavailable or this channel's format is not supported by the effect.
        """
        return SoundEffect(bass_call(BASS_ChannelSetFX, type, priority))

    def bytes_to_seconds(self, position=None):
        """Translates a byte position into time (seconds), based on the format in use by this channel.

        Args:
          position:  The position to translate. Defaults to None

        Returns:
            int: The translated length on success, a negative bass error code on failure.
        """
        position = position or self.position
        return bass_call_0(BASS_ChannelBytes2Seconds, self.handle, position)

    def length_in_seconds(self):
        """Retrieves the length of the stream, in seconds, regardless of position.

        returns:
            int: The length on success, a negative bass error code on failure.
        """
        return self.bytes_to_seconds(self.get_length())

    def seconds_to_bytes(self, position):
        """Translates a time (seconds) position into bytes, based on the format in use by this channel.

        Args:
          position:  The position to translate.

        Returns:
            int: The translated length on success, a negative bass error code on failure.
        """
        return bass_call_0(BASS_ChannelSeconds2Bytes, self.handle, position)

    def get_attribute(self, attribute):
        """Retrieves the value of this channel's attribute.

        Args:
          attribute: The attribute to get the value of. Can either be an of type str or int.

        Returns:
            The value on success, None on failure.

        raises:
            sound_lib.main.BassError: If the attribute is either unavailable or invalid.
                Some attributes have additional possible instances where an exception might be raised.
        """
        value = pointer(c_float())
        if attribute in self.attribute_mapping:
            attribute = self.attribute_mapping[attribute]
        bass_call(BASS_ChannelGetAttribute, self.handle, attribute, value)
        return value.contents.value

    def set_attribute(self, attribute, value):
        """Sets the value of an attribute on this channel.

        Args:
          attribute: The attribute to set the value of. Can either be of type str or int.
          value: 

        Returns:
            bool: True on success, False on failure.

        raises:
            sound_lib.main.BassError: If either attribute or value is invalid.
        """
        if attribute in self.attribute_mapping:
            attribute = self.attribute_mapping[attribute]
        return bass_call(BASS_ChannelSetAttribute, self.handle, attribute, value)

    def slide_attribute(self, attribute, value, time):
        """Slides this channel's attribute from its current value to a new value.

        Args:
          attribute: The attribute to slide the value of.
          value: The new attribute value. Consult specific documentation depending on the one in question.
          time: The length of time (in milliseconds) that it should take for the attribute to reach the value.

        Returns:
            bool: True on success, False on failure.

        raises:
            sound_lib.main.BassError: If attribute is invalid, or the attributes value is set to go from positive to negative or vice versa when the BASS_SLIDE_LOG flag is used.
        """
        if attribute in self.attribute_mapping:
            attribute = self.attribute_mapping[attribute]
        return bass_call(
            BASS_ChannelSlideAttribute, self.handle, attribute, value, time * 1000
        )

    def is_sliding(self, attribute=0):
        """Checks if an attribute (or any attribute) of this channel is sliding. Must be a sample, stream, or MOD music.

        Args:
          attribute: The attribute to check for sliding, or0 for any. Defaults to 0.

        Returns:
            bool: True if sliding, False otherwise.

        """
        return bass_call_0(BASS_ChannelIsSliding, self.handle, attribute)

    def get_info(self):
        """Retrieves information on this channel.

        returns:
            A BASS_CHANNELINFO structure.
        """
        value = pointer(BASS_CHANNELINFO())
        bass_call(BASS_ChannelGetInfo, self.handle, value)
        return value[0]

    def get_level(self):
        """Retrieves the level (peak amplitude) of a stream, MOD music or recording channel.

        returns:
            int: -1 on error. If successful, the level of the left channel is returned in the low word (low 16 bits), and the level of the right channel is returned in the high word (high 16 bits).
                If the channel is mono, then the low word is duplicated in the high word. The level ranges linearly from 0 (silent) to 32768 (max).
                0 will be returned when a channel is stalled.        

        raises:
            sound_lib.main.BassError: If this channel is not playing, or this is a decoding channel which has reached the end
        """
        return bass_call_0(BASS_ChannelGetLevel, self.handle)

    def lock(self):
        """Locks a stream, MOD music or recording channel to the current thread.

        returns:
            bool: True on success, False on failure.
        """
        return bass_call(BASS_ChannelLock, self.handle, True)

    def unlock(self):
        """Unlocks a stream, MOD music or recording channel from the current thread.

        returns:
            bool: True on success, False on failure.
        """
        return bass_call(BASS_ChannelLock, self.handle, False)

    def get_3d_attributes(self):
        """Retrieves the 3D attributes of a sample, stream, or MOD music channel with 3D functionality.

        returns:
            dict: A dict containing the stream's 3d attributes

        raises:
            sound_lib.main.BassError: If this channel does not have 3d functionality.
        """
        answer = dict(
            mode=c_ulong(),
            min=c_float(),
            max=c_float(),
            iangle=c_ulong(),
            oangle=c_ulong(),
            outvol=c_float(),
        )
        bass_call(
            BASS_ChannelGet3DAttributes,
            self.handle,
            pointer(answer["mode"]),
            pointer(answer["min"]),
            pointer(answer["max"]),
            pointer(answer["iangle"]),
            pointer(answer["oangle"]),
            pointer(answer["outvol"]),
        )
        for k in answer:
            answer[k] = answer[k].value()
        return answer

    @update_3d_system
    def set_3d_attributes(
        self, mode=-1, min=0.0, max=0.0, iangle=-1, oangle=-1, outvol=-1
    ):
        """Sets the 3D attributes of a sample, stream, or MOD music channel with 3D functionality.

        Args:
          mode:  The 3D processing mode. Defaults to -1.
          min (float):  The minimum distance. The channel's volume is at maximum when the listener is within this distance... 0 or less = leave current. Defaults to 0.0.
          max (float):  The maximum distance. The channel's volume stops decreasing when the listener is beyond this distance... 0 or less = leave current. Defaults to 0.0.
          iangle (int):  The angle of the inside projection cone in degrees... 0 (no cone) to 360 (sphere), -1 = leave current. Defaults to -1.
          oangle (int):  The angle of the inside projection cone in degrees... 0 (no cone) to 360 (sphere), -1 = leave current. Defaults to -1.
          outvol (float):  The delta-volume outside the outer projection cone... 0 (silent) to 1 (same as inside the cone), less than 0 = leave current. Defaults to -1.0.

        Returns:
            bool: True on success, False otherwise.

        raises:
            sound_lib.main.BassError: If this channel does not have 3d functionality, or one or more attribute values are invalid.
        """
        return bass_call(
            BASS_ChannelSet3DAttributes,
            self.handle,
            mode,
            min,
            max,
            iangle,
            oangle,
            outvol,
        )

    def get_3d_position(self):
        """Retrieves the 3D position of a sample, stream, or MOD music channel with 3D functionality.

        raises:
            sound_lib.main.BassError: If this channel was not initialized with support for 3d functionality.
        """
        answer = dict(
            position=BASS_3DVECTOR(),
            orientation=BASS_3DVECTOR(),
            velocity=BASS_3DVECTOR(),
        )
        bass_call(
            BASS_ChannelGet3DPosition,
            self.handle,
            pointer(answer["position"]),
            pointer(answer["orientation"]),
            pointer(answer["velocity"]),
        )
        return answer

    @update_3d_system
    def set_3d_position(self, position=None, orientation=None, velocity=None):
        """Sets the 3D position of a sample, stream, or MOD music channel with 3D functionality.

        Args:
          position: Defaults to None.
          orientation: Defaults to None.
          velocity: Defaults to None

        raises:
            sound_lib.main.BassError: If this channel was not initialized with support for 3d functionality.
        """
        if position:
            position = pointer(position)
        if orientation:
            orientation = pointer(orientation)
        if velocity:
            velocity = pointer(velocity)
        return bass_call(
            BASS_ChannelSet3DPosition, self.handle, position, orientation, velocity
        )

    def set_link(self, handle):
        """Links two MOD music or stream channels together.

        Args:
          handle: The bass handle to link with this one. Can take both a sound_lib.channel or bass handle. Must be HMUSIC or HSTREAM.

        Returns:
            bool: True on success, False on failure.

        raises:
            sound_lib.main.BassError: If handle points to an invalid channel, either one is a decoding channel, or this channel is already linked to handle.
        """
        if isinstance(handle, Channel):
            handle=handle.handle
        bass_call(BASS_ChannelSetLink, self.handle, handle)

    def remove_link(self, handle):
        """Removes a link between two MOD music or stream channels.

        Args:
          handle: The bass handle to unlink with this one. Can take both a sound_lib.channel or bass handle. Must be a HMUSIC or HSTREAM. Must currently be linked.

        Returns:
            bool: True on success, False on failure.
        raises:
            sound_lib.main.BassError: If chan is either not a valid channel, or is not already linked to handle.

        """
        if isinstance(handle, Channel):
            handle=handle.handle
        return bass_call(BASS_ChannelRemoveLink, self.handle, handle)

    def __iadd__(self, other):
        """Convenience method to link this channel to another.  Calls set_link on the passed in item's handle"""
        self.set_link(other.handle)
        return self

    def __isub__(self, other):
        """Convenience method to unlink this channel from another.  Calls remove_link on the passed in item's handle"""
        self.remove_link(other.handle)
        return self

    def get_frequency(self):
        """Retrieves sample frequency (sample rate).

        returns:
            bool: True on success, False on failure.
        """
        return self.get_attribute(BASS_ATTRIB_FREQ)

    def set_frequency(self, frequency):
        """Sets the frequency (sample rate) of this channel.

        Args:
          frequency (float): The sample rate... 0 = original rate (when the channel was created).

        Returns:
            bool: True on success, False on failure.
        """
        self.set_attribute(BASS_ATTRIB_FREQ, frequency)

    frequency = property(fget=get_frequency, fset=set_frequency)

    def get_pan(self):
        """Gets the panning/balance position of this channel."""
        return self.get_attribute(BASS_ATTRIB_PAN)

    def set_pan(self, pan):
        """Sets the panning/balance position of this channel.

        Args:
          pan (float): The pan position... -1 (full left) to +1 (full right), 0 = centre.

        Returns:
            bool: True on success, False on Failure.

        """
        return self.set_attribute(BASS_ATTRIB_PAN, pan)

    pan = property(fget=get_pan, fset=set_pan)

    def get_volume(self):
        """Gets the volume level of a channel.

        returns:
            float: The volume level... 0 = silent, 1.0 = normal, above 1.0 = amplification.
        """
        return self.get_attribute(BASS_ATTRIB_VOL)

    def set_volume(self, volume):
        """sets the volume level of a channel.

        Args:
          volume (float): The volume level... 0 = silent, 1.0 = normal, above 1.0 = amplification.

        Returns:
            True on success, False on failure.
        """
        self.set_attribute(BASS_ATTRIB_VOL, volume)

    volume = property(fget=get_volume, fset=set_volume)

    def get_data(self, length=16384):
        """Retrieves the immediate sample data (or an FFT representation of it) of this channel. Must be a sample channel, stream, MOD music, or recording channel.

        Args:
          length: Number of bytes wanted (up to 268435455 or 0xFFFFFFF). Defaults to 16384.

        Returns:
            The requested bytes.

        raises:
            sound_lib.main.BassError: If this channel has reached the end, or the BASS_DATA_AVAILABLE flag was used and this is a decoding channel.
        """
        buf = c_buffer(length)
        bass_call_0(BASS_ChannelGetData, self.handle, pointer(buf), length)
        return buf

    # This is less and less of a one-to-one mapping,
    # But I feel that it's better to be consistent with ourselves
    # Than with the library.  We won't punish ourselves
    # For their bad decisions

    def get_looping(self):
        """Returns whether this channel is currently setup to loop."""
        return bass_call_0(BASS_ChannelFlags, self.handle, BASS_SAMPLE_LOOP, 0) == 20

    def set_looping(self, looping):
        """Determines whether this channel is setup to loop.

        Args:
          looping: (bool): Specifies whether this channel should loop.
        """
        if looping:
            return bass_call_0(
                BASS_ChannelFlags, self.handle, BASS_SAMPLE_LOOP, BASS_SAMPLE_LOOP
            )
        return bass_call_0(BASS_ChannelFlags, self.handle, 0, BASS_SAMPLE_LOOP)

    looping = property(fget=get_looping, fset=set_looping)

    def __del__(self):
        try:
            self.free()
        except:
            pass

    def get_x(self):
        """Retrieves this channel's position on the X-axis, if 3d functionality is available.

        raises:
            sound_lib.main.BassError: If this channel was not initialized with support for 3d functionality.
        """
        return self.get_3d_position()["position"].x

    def set_x(self, val):
        """Sets positioning of this channel on the X-axis, if 3d functionality is available.

        Args:
            val: The coordinate position.

        raises:
            sound_lib.main.BassError: If this channel was not initialized with support for 3d functionality.
        """
        pos = self.get_3d_position()
        pos["position"].x = val
        self.set_3d_position(**pos)

    x = property(fget=get_x, fset=set_x)

    def get_y(self):
        """Retrieves this channel's position on the Y-axis, if 3d functionality is available.

        raises:
            sound_lib.main.BassError: If this channel was not initialized with support for 3d functionality.
        """
        return self.get_3d_position()["position"].y

    def set_y(self, val):
        """Sets positioning of this channel on the Y-axis, if 3d functionality is available.

        Args:
          val: The coordinate position.

        raises:
            sound_lib.main.BassError: If this channel was not initialized with support for 3d functionality.
        """
        pos = self.get_3d_position()
        pos["position"].y = val
        self.set_3d_position(**pos)

    y = property(fget=get_y, fset=set_y)

    def get_z(self):
        """Retrieves this channel's position on the Z-axis, if 3d functionality is available.

        raises:
            sound_lib.main.BassError: If this channel was not initialized with support for 3d functionality.
        """
        return self.get_3d_position()["position"].z

    def set_z(self, val):
        """Sets positioning of this channel on the Z-axis, if 3d functionality is available.

        Args:
          val: The coordinate position.

        raises:
            sound_lib.main.BassError: If this channel was not initialized with support for 3d functionality.
        """
        pos = self.get_3d_position()
        pos["position"].z = val
        self.set_3d_position(**pos)

    z = property(fget=get_z, fset=set_z)

    def get_attributes(self):
        """Retrieves all values of all attributes from this object and displays them in a dictionary whose keys are determined by this object's attribute_mapping"""
        res = {}
        for k in self.attribute_mapping:
            try:
                res[k] = self.get_attribute(k)
            except BassError:
                pass
        return res



from ctypes import string_at
import platform
import wave

from .external.pybass import *

if platform.system() == "Windows":
    from .external.pybasswasapi import *
from . import config
from .main import bass_call, bass_call_0


class Input(object):
    """ """
    def __init__(self, device=-1):
        bass_call(BASS_RecordInit, device)
        self._device = device
        self.config = config.BassConfig()

    def free(self):
        """Frees all resources used by the recording device."""
        return bass_call(BASS_RecordFree)

    def get_device(self):
        """ """
        return bass_call_0(BASS_RecordGetDevice)

    def set_device(self, device):
        """

        Args:
          device: 

        Returns:

        """
        if device == self._device:
            return
        self.free()
        self.__init__(device=device)

    device = property(fget=get_device, fset=set_device)

    @staticmethod
    def get_device_names():
        """Convenience method that returns a list of device names that are considered
        	valid by bass.

        Args:

        Returns:
          

        """
        result = ["Default"]
        info = BASS_DEVICEINFO()
        count = 0
        while BASS_RecordGetDeviceInfo(count, ctypes.byref(info)):
            if info.flags & BASS_DEVICE_ENABLED:
                retrieved = info.name
                if platform.system() == "Windows":
                    retrieved = retrieved.decode("mbcs")
                elif platform.system() == "Darwin":
                    retrieved = retrieved.decode("utf-8")
                retrieved = retrieved.replace("(", "").replace(")", "").strip()
                result.append(retrieved)
            count += 1
        return result

    def find_device_by_name(self, name):
        """

        Args:
          name: 

        Returns:

        """
        return self.get_device_names().index(name) - 1

    def find_default_device(self):
        """ """
        return -1

    def find_user_provided_device(self, device_name):
        """

        Args:
          device_name: 

        Returns:

        """
        try:
            return self.find_device_by_name(device_name)
        except ValueError:
            return self.find_default_device()


class WASAPIInput(object):
    """ """
    def __init__(
        self,
        device=-2,
        frequency=0,
        channels=0,
        flags=0,
        buffer=0.0,
        period=0.0,
        callback=None,
    ):
        if callback is None:
            callback = lambda buffer, length, user: True
        self.proc = WASAPIPROC(callback)
        bass_call(
            BASS_WASAPI_Init,
            device,
            frequency,
            channels,
            flags,
            buffer,
            period,
            self.proc,
            None,
        )

    def free(self):
        """ """
        bass_call(BASS_WASAPI_Free)

    def set_device(self, device):
        """

        Args:
          device: 

        Returns:

        """
        bass_call(BASS_WASAPI_SetDevice, device)

    def get_device(self):
        """ """
        return bass_call_0(BASS_WASAPI_GetDevice)

    device = property(fget=get_device, fset=set_device)

    def start(self):
        """ """
        return bass_call(BASS_WASAPI_Start)

    def stop(self, reset=False):
        """

        Args:
          reset:  (Default value = False)

        Returns:

        """
        return bass_call(BASS_WASAPI_Stop, reset)



import os
from platform_utils.paths import module_path, is_frozen, embedded_data_path

if is_frozen():
 x86_path = os.path.join(embedded_data_path(), 'sound_lib', 'lib', 'x86')
 x64_path = os.path.join(embedded_data_path(), 'sound_lib', 'lib', 'x64')
else:
 x86_path = os.path.join(module_path(), '..', 'lib', 'x86')
 x64_path = os.path.join(module_path(), '..', 'lib', 'x64')



from __future__ import absolute_import
from .external.pybass import *
from functools import update_wrapper

EAX_ENVIRONMENTS = {
    "generic": EAX_ENVIRONMENT_GENERIC,
    "padded_cell": EAX_ENVIRONMENT_PADDEDCELL,
    "room": EAX_ENVIRONMENT_ROOM,
    "bathroom": EAX_ENVIRONMENT_BATHROOM,
    "living_room": EAX_ENVIRONMENT_LIVINGROOM,
    "stone_room": EAX_ENVIRONMENT_STONEROOM,
    "auditorium": EAX_ENVIRONMENT_AUDITORIUM,
    "concert_hall": EAX_ENVIRONMENT_CONCERTHALL,
    "cave": EAX_ENVIRONMENT_CAVE,
    "arena": EAX_ENVIRONMENT_ARENA,
    "hanger": EAX_ENVIRONMENT_HANGAR,
    "carpeted_hallway": EAX_ENVIRONMENT_CARPETEDHALLWAY,
    "hallway": EAX_ENVIRONMENT_HALLWAY,
    "stone_corridor": EAX_ENVIRONMENT_STONECORRIDOR,
    "alley": EAX_ENVIRONMENT_ALLEY,
    "forest": EAX_ENVIRONMENT_FOREST,
    "city": EAX_ENVIRONMENT_CITY,
    "mountains": EAX_ENVIRONMENT_MOUNTAINS,
    "quarry": EAX_ENVIRONMENT_QUARRY,
    "plain": EAX_ENVIRONMENT_PLAIN,
    "parking_lot": EAX_ENVIRONMENT_PARKINGLOT,
    "sewer_pipe": EAX_ENVIRONMENT_SEWERPIPE,
    "underwater": EAX_ENVIRONMENT_UNDERWATER,
    "drugged": EAX_ENVIRONMENT_DRUGGED,
    "dizzy": EAX_ENVIRONMENT_DIZZY,
    "psychotic": EAX_ENVIRONMENT_PSYCHOTIC,
}


class BassError(Exception):
    """Error that is raised when there is a problem with a Bass call."""

    def __init__(self, code, description):
        self.code = code
        self.description = description

    def __str__(self):
        return "%d, %s" % (self.code, self.description)


def bass_call(function, *args):
    """Makes a call to bass and raises an exception if it fails.
    This will most likely prove unnecessary for external usage, however we keep it just in case.

    Args:
      function (str): Name of the internal bass function to be called.
      *args: Arguments that will be passed to function.

    Returns:
        Raw output from the specified bass function. Refer to the official docs.

    """
    res = function(*args)
    if res == 0 or res == -1:
        code = BASS_ErrorGetCode()
        raise BassError(code, get_error_description(code))
    return res


def bass_call_0(function, *args):
    """Makes a call to bass and raises an exception if it fails.  Does not consider 0 an error.
    This will most likely prove unnecessary for external usage, however we keep it just in case.

    Args:
      function (str): Name of the internal bass function to be called.
      *args: Arguments that will be passed to function.

    Returns:
        Raw output from the specified bass function. Refer to the official docs.
    """
    res = function(*args)
    if res == -1:
        code = BASS_ErrorGetCode()
        raise BassError(code, get_error_description(code))
    return res


def update_3d_system(func):
    """Decorator to automatically update the 3d system after a function call.
    """

    def update_3d_system_wrapper(*args, **kwargs):
        """

        Args:
          *args: 
          **kwargs: 

        Returns:

        """
        val = func(*args, **kwargs)
        bass_call(BASS_Apply3D)
        return val

    update_wrapper(update_3d_system_wrapper, func)
    return update_3d_system_wrapper


class FlagObject(object):
    """An object which translates bass flags into human-readable/usable items"""
    flag_mapping = {}

    def flags_for(self, **flags):
        """Retrieves flags for given attributes.

        Args:
          **flags: A mapping of human-readable attribute names to numeric values.

        Returns:
            int: A bitmask containing the specified flags, or 0 if nothing was provided.
        """
        res = 0
        for k, v in flags.items():
            if v:
                res |= self.flag_mapping[k]
        return res

    def setup_flag_mapping(self):
        """Sets up a class-level mapping of common flags, in the format human-readable name:value.
        Tipically expanded upon in a subclass.
        """
        self.flag_mapping = {
            "loop": BASS_SAMPLE_LOOP,
            "autofree": BASS_STREAM_AUTOFREE,
            "mono": BASS_SAMPLE_MONO,
            "software": BASS_SAMPLE_SOFTWARE,
            "three_d": BASS_SAMPLE_3D,
            "fx": BASS_SAMPLE_FX,
            "decode": BASS_STREAM_DECODE,
        }



from __future__ import absolute_import
# Copyright(c) Max Kolosov 2009 maxkolosov@inbox.ru
# http://vosolok2008.narod.ru
# BSD license

__version__ = '0.1'
__versionTime__ = '2009-11-15'
__author__ = 'Max Kolosov <maxkolosov@inbox.ru>'
__doc__ = '''
pybassmidi.py - is ctypes python module for
BASSMIDI - extension to the BASS audio library,
enabling the playing of MIDI files and real-time events,
using SF2 soundfonts to provide the sounds.
'''

import sys, ctypes, platform, os
from . import pybass
from .paths import x86_path, x64_path
import libloader

bassmidi_module = libloader.load_library('bassmidi', x86_path=x86_path, x64_path=x64_path)
func_type = libloader.get_functype()
#Register the plugin with the Bass plugin system.
pybass.BASS_PluginLoad(libloader.find_library_path('bassmidi', x86_path=x86_path, x64_path=x64_path), 0)

HSOUNDFONT = ctypes.c_ulong

QWORD = pybass.QWORD
HSTREAM = pybass.HSTREAM
DOWNLOADPROC = pybass.DOWNLOADPROC
BASS_FILEPROCS = pybass.BASS_FILEPROCS


# Additional BASS_SetConfig options
BASS_CONFIG_MIDI_COMPACT = 0x10400
BASS_CONFIG_MIDI_VOICES = 0x10401
BASS_CONFIG_MIDI_AUTOFONT = 0x10402

# Additional BASS_SetConfigPtr options
BASS_CONFIG_MIDI_DEFFONT = 0x10403

# Additional sync types
BASS_SYNC_MIDI_MARKER = 0x10000
BASS_SYNC_MIDI_CUE = 0x10001
BASS_SYNC_MIDI_LYRIC = 0x10002
BASS_SYNC_MIDI_TEXT = 0x10003
BASS_SYNC_MIDI_EVENT = 0x10004
BASS_SYNC_MIDI_TICK = 0x10005
BASS_SYNC_MIDI_TIMESIG = 0x10006

# Additional BASS_MIDI_StreamCreateFile/etc flags
BASS_MIDI_DECAYEND = 0x1000
BASS_MIDI_NOFX = 0x2000
BASS_MIDI_DECAYSEEK = 0x4000

class BASS_MIDI_FONT(ctypes.Structure):
	_fields_ = [('font', HSOUNDFONT),#HSOUNDFONT font; // soundfont
				('preset', ctypes.c_int),#int preset; // preset number (-1=all)
				('bank', ctypes.c_int)#int bank;
				]

class BASS_MIDI_FONTINFO(ctypes.Structure):
	_fields_ = [('name', ctypes.c_char_p),#const char *name;
				('copyright', ctypes.c_char_p),#const char *copyright;
				('comment', ctypes.c_char_p),#const char *comment;
				('presets', ctypes.c_ulong),#DWORD presets; // number of presets/instruments
				('samsize', ctypes.c_ulong),#DWORD samsize; // total size (in bytes) of the sample data
				('samload', ctypes.c_ulong),#DWORD samload; // amount of sample data currently loaded
				('samtype', ctypes.c_ulong)#DWORD samtype; // sample format (CTYPE) if packed
				]

class BASS_MIDI_MARK(ctypes.Structure):
	_fields_ = [('track', ctypes.c_ulong),#DWORD track; // track containing marker
				('pos', ctypes.c_ulong),#DWORD pos; // marker position (bytes)
				('text', ctypes.c_char_p)#const char *text; // marker text
				]

# Marker types
BASS_MIDI_MARK_MARKER = 0 # marker events
BASS_MIDI_MARK_CUE = 1 # cue events
BASS_MIDI_MARK_LYRIC = 2 # lyric events
BASS_MIDI_MARK_TEXT = 3 # text events
BASS_MIDI_MARK_TIMESIG = 4 # time signature

# MIDI events
MIDI_EVENT_NOTE = 1
MIDI_EVENT_PROGRAM = 2
MIDI_EVENT_CHANPRES = 3
MIDI_EVENT_PITCH = 4
MIDI_EVENT_PITCHRANGE = 5
MIDI_EVENT_DRUMS = 6
MIDI_EVENT_FINETUNE = 7
MIDI_EVENT_COARSETUNE = 8
MIDI_EVENT_MASTERVOL = 9
MIDI_EVENT_BANK = 10
MIDI_EVENT_MODULATION = 11
MIDI_EVENT_VOLUME = 12
MIDI_EVENT_PAN = 13
MIDI_EVENT_EXPRESSION = 14
MIDI_EVENT_SUSTAIN = 15
MIDI_EVENT_SOUNDOFF = 16
MIDI_EVENT_RESET = 17
MIDI_EVENT_NOTESOFF = 18
MIDI_EVENT_PORTAMENTO = 19
MIDI_EVENT_PORTATIME = 20
MIDI_EVENT_PORTANOTE = 21
MIDI_EVENT_MODE = 22
MIDI_EVENT_REVERB = 23
MIDI_EVENT_CHORUS = 24
MIDI_EVENT_CUTOFF = 25
MIDI_EVENT_RESONANCE = 26
MIDI_EVENT_RELEASE = 27
MIDI_EVENT_ATTACK = 28
MIDI_EVENT_REVERB_MACRO = 30
MIDI_EVENT_CHORUS_MACRO = 31
MIDI_EVENT_REVERB_TIME = 32
MIDI_EVENT_REVERB_DELAY = 33
MIDI_EVENT_REVERB_LOCUTOFF = 34
MIDI_EVENT_REVERB_HICUTOFF = 35
MIDI_EVENT_REVERB_LEVEL = 36
MIDI_EVENT_CHORUS_DELAY = 37
MIDI_EVENT_CHORUS_DEPTH = 38
MIDI_EVENT_CHORUS_RATE = 39
MIDI_EVENT_CHORUS_FEEDBACK = 40
MIDI_EVENT_CHORUS_LEVEL = 41
MIDI_EVENT_CHORUS_REVERB = 42
MIDI_EVENT_DRUM_FINETUNE = 50
MIDI_EVENT_DRUM_COARSETUNE = 51
MIDI_EVENT_DRUM_PAN = 52
MIDI_EVENT_DRUM_REVERB = 53
MIDI_EVENT_DRUM_CHORUS = 54
MIDI_EVENT_DRUM_CUTOFF = 55
MIDI_EVENT_DRUM_RESONANCE = 56
MIDI_EVENT_DRUM_LEVEL = 57
MIDI_EVENT_TEMPO = 62
MIDI_EVENT_MIXLEVEL = 0x10000
MIDI_EVENT_TRANSPOSE = 0x10001

class BASS_MIDI_EVENT(ctypes.Structure):
	_fields_ = [('event', ctypes.c_ulong),#DWORD event; // MIDI_EVENT_xxx
				('param', ctypes.c_ulong),#DWORD param;
				('chan', ctypes.c_ulong),#DWORD chan;
				('tick', ctypes.c_ulong),#DWORD tick; // event position (ticks)
				('pos', ctypes.c_ulong)#DWORD pos; // event position (bytes)
				]

# BASS_CHANNELINFO type
BASS_CTYPE_STREAM_MIDI = 0x10d00

# Additional attributes
BASS_ATTRIB_MIDI_PPQN = 0x12000
BASS_ATTRIB_MIDI_TRACK_VOL = 0x12100 # + track #

# Additional tag type
BASS_TAG_MIDI_TRACK = 0x11000 # + track #, track text : array of null-terminated ANSI strings

# BASS_ChannelGetLength/GetPosition/SetPosition mode
BASS_POS_MIDI_TICK = 2 # tick position

#HSTREAM BASSMIDIDEF(BASS_MIDI_StreamCreate)(DWORD channels, DWORD flags, DWORD freq);
BASS_MIDI_StreamCreate = func_type(HSTREAM, ctypes.c_ulong, ctypes.c_ulong, ctypes.c_ulong)(('BASS_MIDI_StreamCreate', bassmidi_module))
#HSTREAM BASSMIDIDEF(BASS_MIDI_StreamCreateFile)(BOOL mem, const void *file, QWORD offset, QWORD length, DWORD flags, DWORD freq);
BASS_MIDI_StreamCreateFile = func_type(HSTREAM, ctypes.c_byte, ctypes.c_void_p, QWORD, QWORD, ctypes.c_ulong, ctypes.c_ulong)(('BASS_MIDI_StreamCreateFile', bassmidi_module))
#HSTREAM BASSMIDIDEF(BASS_MIDI_StreamCreateURL)(const char *url, DWORD offset, DWORD flags, DOWNLOADPROC *proc, void *user, DWORD freq);
BASS_MIDI_StreamCreateURL = func_type(HSTREAM, ctypes.c_char_p, ctypes.c_ulong, ctypes.c_ulong, DOWNLOADPROC, ctypes.c_void_p, ctypes.c_ulong)(('BASS_MIDI_StreamCreateURL', bassmidi_module))
#HSTREAM BASSMIDIDEF(BASS_MIDI_StreamCreateFileUser)(DWORD system, DWORD flags, const BASS_FILEPROCS *procs, void *user, DWORD freq);
BASS_MIDI_StreamCreateFileUser = func_type(HSTREAM, ctypes.c_ulong, ctypes.c_ulong, ctypes.POINTER(BASS_FILEPROCS), ctypes.c_void_p, ctypes.c_ulong)(('BASS_MIDI_StreamCreateFileUser', bassmidi_module))
#BOOL BASSMIDIDEF(BASS_MIDI_StreamGetMark)(HSTREAM handle, DWORD type, DWORD index, BASS_MIDI_MARK *mark);
BASS_MIDI_StreamGetMark = func_type(ctypes.c_byte, HSTREAM, ctypes.c_ulong, ctypes.c_ulong, ctypes.POINTER(BASS_MIDI_MARK))(('BASS_MIDI_StreamGetMark', bassmidi_module))
#BOOL BASSMIDIDEF(BASS_MIDI_StreamSetFonts)(HSTREAM handle, const BASS_MIDI_FONT *fonts, DWORD count);
BASS_MIDI_StreamSetFonts = func_type(ctypes.c_byte, HSTREAM, ctypes.POINTER(BASS_MIDI_FONT), ctypes.c_ulong)(('BASS_MIDI_StreamSetFonts', bassmidi_module))
#DWORD BASSMIDIDEF(BASS_MIDI_StreamGetFonts)(HSTREAM handle, BASS_MIDI_FONT *fonts, DWORD count);
BASS_MIDI_StreamGetFonts = func_type(ctypes.c_ulong, HSTREAM, ctypes.POINTER(BASS_MIDI_FONT), ctypes.c_ulong)(('BASS_MIDI_StreamGetFonts', bassmidi_module))
#BOOL BASSMIDIDEF(BASS_MIDI_StreamLoadSamples)(HSTREAM handle);
BASS_MIDI_StreamLoadSamples = func_type(ctypes.c_byte, HSTREAM)(('BASS_MIDI_StreamLoadSamples', bassmidi_module))
#BOOL BASSMIDIDEF(BASS_MIDI_StreamEvent)(HSTREAM handle, DWORD chan, DWORD event, DWORD param);
BASS_MIDI_StreamEvent = func_type(ctypes.c_byte, HSTREAM, ctypes.c_ulong, ctypes.c_ulong, ctypes.c_ulong)(('BASS_MIDI_StreamEvent', bassmidi_module))
#DWORD BASSMIDIDEF(BASS_MIDI_StreamGetEvent)(HSTREAM handle, DWORD chan, DWORD event);
BASS_MIDI_StreamGetEvent = func_type(ctypes.c_ulong, HSTREAM, ctypes.c_ulong, ctypes.c_ulong)(('BASS_MIDI_StreamGetEvent', bassmidi_module))
#DWORD BASSMIDIDEF(BASS_MIDI_StreamGetEvents)(HSTREAM handle, DWORD track, DWORD filter, BASS_MIDI_EVENT *events);
BASS_MIDI_StreamGetEvents = func_type(ctypes.c_ulong, HSTREAM, ctypes.c_ulong, ctypes.c_ulong, ctypes.POINTER(BASS_MIDI_EVENT))(('BASS_MIDI_StreamGetEvents', bassmidi_module))
#HSTREAM BASSMIDIDEF(BASS_MIDI_StreamGetChannel)(HSTREAM handle, DWORD chan);
BASS_MIDI_StreamGetChannel = func_type(HSTREAM, HSTREAM, ctypes.c_ulong)(('BASS_MIDI_StreamGetChannel', bassmidi_module))

#HSOUNDFONT BASSMIDIDEF(BASS_MIDI_FontInit)(const void *file, DWORD flags);
BASS_MIDI_FontInit = func_type(HSOUNDFONT, ctypes.c_void_p, ctypes.c_ulong)(('BASS_MIDI_FontInit', bassmidi_module))
#BOOL BASSMIDIDEF(BASS_MIDI_FontFree)(HSOUNDFONT handle);
BASS_MIDI_FontFree = func_type(ctypes.c_byte, HSOUNDFONT)(('BASS_MIDI_FontFree', bassmidi_module))
#BOOL BASSMIDIDEF(BASS_MIDI_FontGetInfo)(HSOUNDFONT handle, BASS_MIDI_FONTINFO *info);
BASS_MIDI_FontGetInfo = func_type(ctypes.c_byte, HSOUNDFONT, ctypes.POINTER(BASS_MIDI_FONTINFO))(('BASS_MIDI_FontGetInfo', bassmidi_module))
#const char *BASSMIDIDEF(BASS_MIDI_FontGetPreset)(HSOUNDFONT handle, int preset, int bank);
BASS_MIDI_FontGetPreset = func_type(ctypes.c_char_p, HSOUNDFONT, ctypes.c_int, ctypes.c_int)(('BASS_MIDI_FontGetPreset', bassmidi_module))
#BOOL BASSMIDIDEF(BASS_MIDI_FontLoad)(HSOUNDFONT handle, int preset, int bank);
BASS_MIDI_FontLoad = func_type(ctypes.c_byte, HSOUNDFONT, ctypes.c_int, ctypes.c_int)(('BASS_MIDI_FontLoad', bassmidi_module))
#BOOL BASSMIDIDEF(BASS_MIDI_FontCompact)(HSOUNDFONT handle);
BASS_MIDI_FontCompact = func_type(ctypes.c_byte, HSOUNDFONT)(('BASS_MIDI_FontCompact', bassmidi_module))
#BOOL BASSMIDIDEF(BASS_MIDI_FontPack)(HSOUNDFONT handle, const void *outfile, const void *encoder, DWORD flags);
BASS_MIDI_FontPack = func_type(ctypes.c_byte, HSOUNDFONT, ctypes.c_void_p, ctypes.c_void_p, ctypes.c_ulong)(('BASS_MIDI_FontPack', bassmidi_module))
#BOOL BASSMIDIDEF(BASS_MIDI_FontUnpack)(HSOUNDFONT handle, const void *outfile, DWORD flags);
BASS_MIDI_FontUnpack = func_type(ctypes.c_byte, HSOUNDFONT, ctypes.c_void_p, ctypes.c_ulong)(('BASS_MIDI_FontUnpack', bassmidi_module))
#BOOL BASSMIDIDEF(BASS_MIDI_FontSetVolume)(HSOUNDFONT handle, float volume);
BASS_MIDI_FontSetVolume = func_type(ctypes.c_byte, HSOUNDFONT, ctypes.c_float)(('BASS_MIDI_FontSetVolume', bassmidi_module))
#float BASSMIDIDEF(BASS_MIDI_FontGetVolume)(HSOUNDFONT handle);
BASS_MIDI_FontGetVolume = func_type(ctypes.c_float, HSOUNDFONT)(('BASS_MIDI_FontGetVolume', bassmidi_module))




from __future__ import absolute_import
from .channel import Channel
from .external.pybass import *
from ctypes import string_at
import wave
from .main import bass_call, bass_call_0


class Recording(Channel):
    """ """
    def __init__(
        self, frequency=44100, channels=2, flags=BASS_RECORD_PAUSE, proc=None, user=None
    ):
        if not proc:
            proc = lambda: True
        self.callback = RECORDPROC(proc)
        self._frequency = frequency
        self._channels = channels
        self._flags = flags
        handle = bass_call(
            BASS_RecordStart, frequency, channels, flags, self.callback, user
        )
        super(Recording, self).__init__(handle)

    def free(self):
        """ """
        pass


class WaveRecording(Recording):
    """ """
    def __init__(self, filename=None, proc=None, *args, **kwargs):
        callback = proc or self.recording_callback
        super(WaveRecording, self).__init__(proc=callback, *args, **kwargs)
        self.filename = filename

    def recording_callback(self, handle, buffer, length, user):
        """

        Args:
          handle: 
          buffer: 
          length: 
          user: 

        Returns:

        """
        buf = string_at(buffer, length)
        self.file.writeframes(buf)
        return True

    def setup_file(self):
        """ """
        self.file = wave.open(self.filename, "w")
        self.file.setnchannels(self._channels)
        self.file.setsampwidth(2)
        self.file.setframerate(self._frequency)

    def play(self, *args, **kwargs):
        """

        Args:
          *args: 
          **kwargs: 

        Returns:

        """
        if not self.is_playing:
            self.setup_file()
        super(WaveRecording, self).play(*args, **kwargs)

    def stop(self, *args, **kwargs):
        """

        Args:
          *args: 
          **kwargs: 

        Returns:

        """
        super(WaveRecording, self).stop(*args, **kwargs)
        self.file.close()



from __future__ import absolute_import
from .external import pybass, pybassenc
from .main import bass_call, bass_call_0, FlagObject


class Encoder(FlagObject):
    """ """
    def setup_flag_mapping(self):
        """ """
        # super(Encoder, self).setup_flag_mapping()
        self.flag_mapping = {
            "pcm": pybassenc.BASS_ENCODE_PCM,
            "no_header": pybassenc.BASS_ENCODE_NOHEAD,
            "rf64": pybassenc.BASS_ENCODE_RF64,
            "big_endian": pybassenc.BASS_ENCODE_BIGEND,
            "fp_8bit": pybassenc.BASS_ENCODE_FP_8BIT,
            "fp_16bit": pybassenc.BASS_ENCODE_FP_16BIT,
            "fp_24bit": pybassenc.BASS_ENCODE_FP_24BIT,
            "fp_32bit": pybassenc.BASS_ENCODE_FP_32BIT,
            "queue": pybassenc.BASS_ENCODE_QUEUE,
            "limit": pybassenc.BASS_ENCODE_LIMIT,
            "no_limit": pybassenc.BASS_ENCODE_CAST_NOLIMIT,
            "pause": pybassenc.BASS_ENCODE_PAUSE,
            "autofree": pybassenc.BASS_ENCODE_AUTOFREE,
            "unicode": pybass.BASS_UNICODE,
        }

    def __init__(
        self,
        source,
        command_line,
        pcm=False,
        no_header=False,
        rf64=False,
        big_endian=False,
        fp_8bit=False,
        fp_16bit=False,
        fp_24bit=False,
        fp_32bit=False,
        queue=False,
        limit=False,
        no_limit=False,
        pause=True,
        autofree=False,
        callback=None,
        user=None,
    ):
        self.setup_flag_mapping()
        flags = self.flags_for(
            pcm=pcm,
            no_header=no_header,
            rf64=rf64,
            big_endian=big_endian,
            fp_8bit=fp_8bit,
            fp_16bit=fp_16bit,
            fp_24bit=fp_24bit,
            fp_32bit=fp_32bit,
            queue=queue,
            limit=limit,
            no_limit=no_limit,
            pause=pause,
            autofree=autofree,
        )  # fwiw!
        self.source = source
        source_handle = source.handle
        if callback is None:
            callback = lambda *a: None
        callback = pybassenc.ENCODEPROC(callback)
        self.callback = callback
        self.handle = bass_call(
            pybassenc.BASS_Encode_Start,
            source_handle,
            command_line,
            flags,
            callback,
            user,
        )

    @property
    def paused(self):
        """ """
        return (
            bass_call_0(pybassenc.BASS_Encode_IsActive, self.handle)
            == pybass.BASS_ACTIVE_PAUSED
        )

    @paused.setter
    def paused(self, paused):
        """

        Args:
          paused: 

        Returns:

        """
        return bass_call(pybassenc.BASS_Encode_SetPaused, self.handle, paused)

    def is_stopped(self):
        """ """
        return (
            bass_call_0(pybassenc.BASS_Encode_IsActive, self.handle)
            == pybass.BASS_ACTIVE_STOPPED
        )

    def stop(self):
        """ """
        return bass_call(pybassenc.BASS_Encode_Stop, self.handle)


class BroadcastEncoder(Encoder):
    """ """
    def __init__(
        self,
        source_encoder,
        server,
        password,
        content,
        name=None,
        url=None,
        genre=None,
        description=None,
        headers=None,
        bitrate=0,
        public=False,
    ):
        contents = {
            "mp3": pybassenc.BASS_ENCODE_TYPE_MP3,
            "ogg": pybassenc.BASS_ENCODE_TYPE_OGG,
            "aac": pybassenc.BASS_ENCODE_TYPE_AAC,
        }
        if content in contents:
            content = contents[content]
        self.source_encoder = source_encoder
        handle = source_encoder.handle
        self.server = server
        self.password = password
        self.status = bass_call(
            pybassenc.BASS_Encode_CastInit,
            handle,
            server,
            password,
            content,
            name,
            url,
            genre,
            description,
            headers,
            bitrate,
            public,
        )

    def set_title(self, title=None, url=None):
        """

        Args:
          title:  (Default value = None)
          url:  (Default value = None)

        Returns:

        """
        return bass_call(
            pybassenc.BASS_Encode_CastSetTitle, self.source_encoder.handle, title, url
        )

    def get_stats(self, type, password=None):
        """

        Args:
          type: 
          password:  (Default value = None)

        Returns:

        """
        types = {
            "shoutcast": pybassenc.BASS_ENCODE_STATS_SHOUT,
            "icecast": pybassenc.BASS_ENCODE_STATS_ICE,
            "icecast_server": pybassenc.BASS_ENCODE_STATS_ICESERV,
        }
        if type in types:
            type = types[type]
        if password is None:
            password = self.password
        return bass_call(
            pybassenc.BASS_Encode_CastGetStats, self.handle, type, password
        )



from __future__ import absolute_import
# Copyright(c) Max Kolosov 2009 maxkolosov@inbox.ru
# http://vosolok2008.narod.ru
# BSD license

__version__ = '0.1'
__versionTime__ = '2009-11-10'
__author__ = 'Max Kolosov <maxkolosov@inbox.ru>'
__doc__ = '''
pytags.py - is ctypes python module for
TAGS (Another Tags Reading Library written by Wraith).

BASS audio library has limited support for reading tags, associated with a
stream. This library extends that functionality, allowing developer/user to
extract specific song information from the stream handle used with BASS. The
extracted tag values are formatted into text ouput according to given format
string (including conditional processing).

Supported tags:
---------------
  MP3 ID3v1 and ID3v2.2/3/4
  OGG/FLAC comments
  WMA
  APE, OFR, MPC, AAC - all use APE tags
  MP4
  MOD/etc titles
'''

import sys, ctypes, platform
from .paths import x86_path, x64_path
import libloader
tags_module = libloader.load_library('tags', x86_path=x86_path, x64_path=x64_path)
func_type = libloader.get_functype()# Current version. Just increments each release.

TAGS_VERSION = 17

# returns description of the last error.
#const char*  _stdcall TAGS_GetLastErrorDesc();
TAGS_GetLastErrorDesc = func_type(ctypes.c_char_p)(('TAGS_GetLastErrorDesc', tags_module))

# main purpose of this library
#const char*  _stdcall TAGS_Read( DWORD dwHandle, const char* fmt );
TAGS_Read = func_type(ctypes.c_char_p, ctypes.c_ulong, ctypes.c_char_p)(('TAGS_Read', tags_module))

# retrieves the current version
#DWORD _stdcall TAGS_GetVersion();
TAGS_GetVersion = func_type(ctypes.c_ulong)(('TAGS_GetVersion', tags_module))

